/*
 * C
 *
 * Copyright 2017-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

//#define NET_ADAPT_PRINT_ERROR
//#define NET_ADAPT_DEBUG

#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <sni.h>

#include "LLNET_CONSTANTS.h"
#include "LLNET_ERRORS.h"
#include "NETIF_interface.h"
#include "pool.h"
#include "NET_ADAPT_delegate_task.h"
#include "WIFI_WF200_driver.h"

#include "fsl_debug_console.h"

NET_ADAPT_status_t NET_ADAPT_init_f(void)
{
	NET_ADAPT_status_t e_return = NET_ADAPT_NO_ERROR;

	return e_return;
}

NET_ADAPT_status_t NET_ADAPT_get_java_thread(unsigned long * const _pul_java_thread)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	NET_ADAPT_status_t e_return;
	int32_t i32_java_thread;
	i32_java_thread = SNI_getCurrentJavaThreadID();

	if ((_pul_java_thread != NULL) && (i32_java_thread != SNI_ERROR))
	{
		* _pul_java_thread = (unsigned long)i32_java_thread;
		e_return = NET_ADAPT_NO_ERROR;
	}
	else
	{
		#ifdef NET_ADAPT_PRINT_ERROR
			PRINTF("NET_ADAPT_ERROR [%s:%d] : cannot get java thread (error %d)\n", __FILE__, __LINE__, i32_java_thread);
		#endif
		e_return = NET_ADAPT_ERROR;
	}

	return (e_return);
}

NET_ADAPT_status_t NET_ADAPT_suspend_java_thread (void)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	NET_ADAPT_status_t e_return = NET_ADAPT_NO_ERROR;
	int32_t i32_result;

	i32_result = SNI_suspendCurrentJavaThread(0);

	if (i32_result != SNI_OK)
	{
		e_return = NET_ADAPT_ERROR;
		#ifdef NET_ADAPT_PRINT_ERROR
			PRINTF("NET_ADAPT_ERROR [%s:%d] : cannot suspend java thread (error %d)\n", __FILE__, __LINE__, i32_result);
		#endif
	}

	return e_return;
}

NET_ADAPT_status_t NET_ADAPT_resume_java_thread(const unsigned long _ul_java_thread)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	NET_ADAPT_status_t e_return = NET_ADAPT_NO_ERROR;
	int32_t i32_result;

	i32_result = SNI_resumeJavaThread(_ul_java_thread);

	if (i32_result != SNI_OK)
	{
		#ifdef NET_ADAPT_PRINT_ERROR
			PRINTF("NET_ADAPT_ERROR [%s:%d] : cannot resume java thread (error %d, thread %lu)\n", __FILE__, __LINE__, i32_result, _ul_java_thread);
		#endif
		e_return = NET_ADAPT_ERROR;
	}

	return e_return;
}

NET_ADAPT_status_t NET_ADAPT_post_event_action_f(void * const _pv_mailbox, void * const _pv_event, const unsigned long _ul_timeout)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	NET_ADAPT_status_t e_return = NET_ADAPT_NO_ERROR;
	BaseType_t e_result;

	/* test if mailbox is full */
	if (uxQueueSpacesAvailable(*(QueueHandle_t*)(_pv_mailbox)) != 0)
	{
		/* send event without timeout */
		e_result = xQueueSend(*(QueueHandle_t*)(_pv_mailbox), &_pv_event, 0);

		/* test if send is nok */
		if (e_result != pdPASS)
		{
			e_return = NET_ADAPT_ERROR;
			#ifdef NET_ADAPT_PRINT_ERROR
				PRINTF("NET_ADAPT_ERROR [%s:%d] : sending error (err %u)\n", __FILE__, __LINE__, e_result);
			#endif
		}
	}
	else
	{
		e_return = NET_ADAPT_ERROR;
		#ifdef NET_ADAPT_PRINT_ERROR
			PRINTF("NET_ADAPT_ERROR [%s:%d] : mail box appear full\n", __FILE__, __LINE__);
		#endif
	}

	return e_return;
}

NET_ADAPT_status_t NET_ADAPT_wait_event(void * const _pv_mailbox, void ** const _ppv_event, const unsigned long _ul_timeout)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	NET_ADAPT_status_t e_return = NET_ADAPT_NO_ERROR;
	BaseType_t e_result;

	/* receive event from mailbox */
	e_result = xQueueReceive(*(QueueHandle_t*)(_pv_mailbox), _ppv_event, (TickType_t)(_ul_timeout / portTICK_PERIOD_MS));
	/* test received function */
	if (e_result != pdTRUE)
	{
		e_return = NET_ADAPT_ERROR;
		#ifdef NET_ADAPT_PRINT_ERROR
			PRINTF("NET_ADAPT_ERROR [%s:%d] : task was wake up not properly\n", __FILE__, __LINE__);
		#endif
	}

	return e_return;
}

///////////////// ECOM-NETWORK START ////////////////////

NET_ADAPT_status_t NET_ADAPT_start_f(char * const _pc_interface_name_in,
																			int _i_interface_name_length_in,
																			short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_pc_interface_name_in != NULL);
	configASSERT(_i_interface_name_length_in > 0);
	configASSERT(_ps_error_out != NULL);

	/* generic start interface */
	if (NETIF_start_f(_pc_interface_name_in) != NETIF_NO_ERROR)
	{
		*_ps_error_out = J_EUNKNOWN;
	}
	else
	{
		*_ps_error_out = 0;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_stop_f(char * const _pc_interface_name_in,
																		int _i_interface_name_length_in,
                                    short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_pc_interface_name_in != NULL);
	configASSERT(_i_interface_name_length_in > 0);
	configASSERT(_ps_error_out != NULL);

	/* generic stop interface */
	if (NETIF_stop_f(_pc_interface_name_in) != NETIF_NO_ERROR)
	{
		*_ps_error_out = J_EUNKNOWN;
	}
	else
	{
		*_ps_error_out = 0;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_apply_configuration_f(char * const _pc_interface_name_in,
													int _i_interface_name_length_in,
													short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_pc_interface_name_in != NULL);
	configASSERT(_i_interface_name_length_in > 0);
	configASSERT(_ps_error_out != NULL);

	/* generic stop interface */
	if (NETIF_apply_configuration_f(_pc_interface_name_in) != NETIF_NO_ERROR)
	{
		*_ps_error_out = J_EUNKNOWN;
	}
	else
	{
		*_ps_error_out = 0;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_disable_soft_ap_f(short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_disable_softap_f())
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_enable_soft_ap_f(char * const _pc_ssid_in,
													int _i_ssid_length_in,
													char * const _pc_passphrase_in,
													int _i_passphrase_length_in,
													short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);
	configASSERT(_pc_ssid_in != NULL);
	configASSERT(_i_ssid_length_in > 0);
	configASSERT(_pc_passphrase_in != NULL);
	configASSERT(_i_passphrase_length_in > 0);

	if (WIFI_WF200_enable_softap_f( _pc_ssid_in, _i_ssid_length_in, _pc_passphrase_in, _i_passphrase_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_set_name_soft_ap_f(char * const _pc_name_in,
													int _i_name_length_in,
													short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_pc_name_in != NULL);
	configASSERT(_i_name_length_in >= 0);
	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_set_name_softap_f( _pc_name_in, _i_name_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_bssid_f(char * const _pc_bssid_in,
												int _i_bssid_length_in,
												short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);
	configASSERT(_pc_bssid_in != NULL);
	configASSERT(_i_bssid_length_in > 0);

	if (WIFI_WF200_get_bssid_f((unsigned char*) _pc_bssid_in, _i_bssid_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_channel_f(short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (!WIFI_WF200_get_channel_f(_ps_error_out))
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_rssi_f(float * const _pf_rssi_in,
											int _i_rssi_length_in,
											short * const _ps_error_out)
{
	configASSERT(_ps_error_out != NULL);
	configASSERT(_pf_rssi_in != NULL);
	configASSERT(_i_rssi_length_in > 0);

	if (WIFI_WF200_get_rssi_f(_pf_rssi_in, _i_rssi_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_security_mode_f(short * const _ps_error_out)
{
	configASSERT(_ps_error_out != NULL);

	if (!WIFI_WF200_get_security_mode_f(_ps_error_out))
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ssid_f(char * const _pc_ssid_in,
											int _i_ssid_length_in,
											short * const _ps_error_out)
{
	configASSERT(_ps_error_out != NULL);
	configASSERT(_pc_ssid_in != NULL);
	configASSERT(_i_ssid_length_in > 0);

	if (WIFI_WF200_get_ssid_f(_pc_ssid_in, _i_ssid_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_wps_modes_f(short * const _ps_error_out)
{
	configASSERT(_ps_error_out != NULL);

	if (!WIFI_WF200_get_wps_modes_f(_ps_error_out))
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_join_f(char * const _pc_ssid_in,
											int _i_ssid_length_in,
											char * const _pc_passphrase_in,
											int _i_passphrase_length_in,
											int _i_security_mode_in,
											short * const _ps_error_out)
{
	configASSERT(_ps_error_out != NULL);
	configASSERT(_pc_ssid_in != NULL);
	configASSERT(_i_ssid_length_in > 0);
	configASSERT(_pc_passphrase_in != NULL);
	configASSERT(_i_passphrase_length_in > 0);
	configASSERT(_i_security_mode_in >= 0);

#ifdef READ_WIFI_CONF_FROM_SDCARD // This is defined in IAR preprocessor options.
  // These externs are temporary to get SSID/PASSWD and Security setting from 
  // wifi_cfg.txt file on SD Card.
  extern char g_wifi_SSID[];
  extern char g_wifi_PASSWD[];
  extern char g_wifi_SECURITY[];

  if ( ( strlen(g_wifi_SSID) > 0 ) && 
       ( strlen(g_wifi_PASSWD) > 0 )  )
  {
		// SSID, Password and security is defined in sdcard, therefore override
		// values requested by application.
		uint32_t sdcard_ssid_length = strlen(g_wifi_SSID);
		uint16_t sdcard_passkey_length = strlen(g_wifi_PASSWD);
		uint32_t sdcard_security_mode = 5;

		if (  strlen(g_wifi_SECURITY) > 0 ) 
		{
		
			if ( strcmp (g_wifi_SECURITY, "OPEN") == 0 )
			{
				sdcard_security_mode = 6;
			} 
			else if ( strcmp (g_wifi_SECURITY, "WEP") == 0  )
			{
				sdcard_security_mode = 7;  // it is 7 or 8
			}    
			else if ( strcmp (g_wifi_SECURITY, "WPA2_WPA1_PSK") == 0   )
			{
				sdcard_security_mode = 5;
			}      
			else if ( strcmp (g_wifi_SECURITY, "WPA2_PSK") == 0   )
			{
				sdcard_security_mode = 10;
			}         
		
		}

        PRINTF ("Joining SD card defined SSID=%s\n" , g_wifi_SSID);

		if (WIFI_WF200_join_f	(g_wifi_SSID, strlen(g_wifi_SSID),
								g_wifi_PASSWD, strlen(g_wifi_PASSWD), 
								sdcard_security_mode))
		{
			*_ps_error_out = 0;
		}
		else
		{
			*_ps_error_out = J_EUNKNOWN;
		}

  }
  else
#endif
  {

#ifdef NET_ADAPT_DEBUG
    PRINTF ("Joining Java app defined SSID=%s\n", _pc_ssid_in );
#endif

	if (WIFI_WF200_join_f(_pc_ssid_in, _i_ssid_length_in, _pc_passphrase_in, _i_passphrase_length_in, _i_security_mode_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}
  }

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_leave_f(short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_leave_f())
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return (NET_ADAPT_NO_ERROR);
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_count_f(short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (!WIFI_WF200_get_ap_count_f(_ps_error_out))
	{
		*_ps_error_out = J_EACCES;
	}

	return NET_ADAPT_NO_ERROR;
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_ssid_f(int _i_index_in,
												unsigned char * const _puc_ssid_in,
												int _i_ssid_length_in,
												short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_get_ap_ssid_f(_i_index_in, _puc_ssid_in, _i_ssid_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return NET_ADAPT_NO_ERROR;
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_bssid_f(int _i_index_in,
													unsigned char * const _puc_bssid_in,
													int _i_bssid_length_in,
													short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_get_ap_bssid_f(_i_index_in, _puc_bssid_in, _i_bssid_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return NET_ADAPT_NO_ERROR;
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_channel_f(int _i_index_in,
													short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif
	int32_t i_channel;

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_get_ap_channel_f(_i_index_in, (int*)&i_channel))
	{
		*_ps_error_out = i_channel;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return NET_ADAPT_NO_ERROR;
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_rssi_f(int _i_index_in,
												float * const _pf_rssi_in,
												int _i_rssi_length_in,
												short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_get_ap_rssi_f(_i_index_in, _pf_rssi_in, _i_rssi_length_in))
	{
		*_ps_error_out = 0;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return NET_ADAPT_NO_ERROR;
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_security_mode_f(int _i_index_in,
															short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	int32_t i_security_mode;

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_get_ap_security_mode_f(_i_index_in, (int*)&i_security_mode))
	{
		*_ps_error_out = i_security_mode;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return NET_ADAPT_NO_ERROR;
}

NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_wps_modes_f(int _i_index_in,
														short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	int32_t i_wps_modes;

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_get_ap_wps_modes_f(_i_index_in, (int*)&i_wps_modes))
	{
		*_ps_error_out = i_wps_modes;
	}
	else
	{
		*_ps_error_out = J_EUNKNOWN;
	}

	return NET_ADAPT_NO_ERROR;
}

NET_ADAPT_status_t NET_ADAPT_get_client_state_f(short * const _ps_error_out)
{
#ifdef NET_ADAPT_DEBUG
	PRINTF("NET_ADAPT_DEBUG [%s:%d]\n", __FILE__, __LINE__);
#endif

	configASSERT(_ps_error_out != NULL);

	if (WIFI_WF200_get_client_state_f())
	{
		/* mean client is started */
		*_ps_error_out = 1;
	}
	else
	{
		*_ps_error_out = 0;
	}

	return NET_ADAPT_NO_ERROR;
}
