/*
 * C
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
#ifndef _DISPLAY_FRAMEBUFFER_H_
#define _DISPLAY_FRAMEBUFFER_H_

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include <stdint.h>

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/*
 * @brief Width of the display panel
 */
#define DISPLAY_PANEL_WIDTH (400)

/*
 * @brief Height of the display panel
 */
#define DISPLAY_PANEL_HEIGHT (392)

/*
 * @brief Width of the frame buffers
 */
#define FRAME_BUFFER_WIDTH (392)

/*
 * @brief Height of the frame buffers
 */
#define FRAME_BUFFER_HEIGHT (392)

/*
 * @brief Available number of bytes per pixels (RGB565)
 */
#define FRAME_BUFFER_BYTE_PER_PIXEL (2)

/*
 * @brief Frame buffer alignment. 
 * - 4 bits are required by SMARTDMA
 * - 64 bits are required by DMA
 * Note: Full frame buffer memory size is a value aligned on 64 too;
 * so next allocation behind frame buffer memory is aligned on 64 too.
 */
#define FRAME_BUFFER_ALIGN (64)

/*
 * @brief Source image line alignment (in bytes), required by VGLite library
 * when an image is used as source: 32 bits for RGB565 format
 * (see _check_source_aligned() in vg_lite.c).
 */
#define VGLITE_IMAGE_LINE_ALIGN_BYTE (32)
#define ALIGN(value, align) (((value) + (align) - 1) & ~((align) - 1))

/*
 * @brief Frame buffer line alignment (in bytes), required by VGLite library
 * when frame buffer is used as source.
 * Note: Set (1) to not use the frame buffer as source image.
 * Note 2: Require a patch in fsl_dc_fb_dsi_cmd.c
 */
#define FRAME_BUFFER_LINE_ALIGN_BYTE (1) /* 1 | VGLITE_IMAGE_LINE_ALIGN_BYTE */

/*
 * @brief Frame buffer line size in bytes (stride), according FRAME_BUFFER_LINE_ALIGN
 */
#define FRAME_BUFFER_STRIDE_BYTE ALIGN(FRAME_BUFFER_WIDTH * FRAME_BUFFER_BYTE_PER_PIXEL, FRAME_BUFFER_LINE_ALIGN_BYTE)

/*
 * @brief Frame buffer line size in pixels (stride), according FRAME_BUFFER_LINE_ALIGN
 */
#define FRAME_BUFFER_STRIDE_PIXELS (FRAME_BUFFER_STRIDE_BYTE / FRAME_BUFFER_BYTE_PER_PIXEL)

/*
 * @brief Available number of frame buffers
 */
#define FRAME_BUFFER_COUNT 2

// -----------------------------------------------------------------------------
// Types
// -----------------------------------------------------------------------------

/*
 * @brief Frame buffer structure definition
 */
typedef struct s_framebuffer {
	uint16_t p[FRAME_BUFFER_HEIGHT][FRAME_BUFFER_STRIDE_PIXELS];
} framebuffer_t;

// -----------------------------------------------------------------------------
// Variables
// -----------------------------------------------------------------------------

/*
 * @brief Frame buffer address array
 */
extern const framebuffer_t * s_frameBufferAddress[FRAME_BUFFER_COUNT];

// -----------------------------------------------------------------------------
// Inlined functions
// -----------------------------------------------------------------------------

/*
 * @brief gets the width of the frame buffers
 *
 * @return: The width of the frame buffers
 */
static inline int display_framebuffer_get_width() { return FRAME_BUFFER_WIDTH; }

/*
 * @brief gets the height of the frame buffers
 *
 * @return: The height of the frame buffers
 */
static inline int display_framebuffer_get_height() { return FRAME_BUFFER_HEIGHT; }

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

#endif	// #ifndef _DISPLAY_FRAMEBUFFER_H_

