/*
 * C
 *
 * Copyright 2019-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

#if !defined DISPLAY_VGLITE_H
#define DISPLAY_VGLITE_H

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include "LLUI_PAINTER_impl.h"

#include "vglite_window.h"

#include "trace_vglite.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Public functions
// -----------------------------------------------------------------------------

/*
 * @brief Notify display_vglite module when a vglite operation has been started
 */
void display_vglite_start_operation(void);

/*
 * @brief Notify display_vglite that a vglite operation has been canceled
 */
void display_vglite_cancel_operation(void);

/*
 * @brief Waits for a vglite operation to complete
 */
void display_vglite_wait_operation(void);

/*
 * @brief Enables hardware rendering
 */
void display_vglite_enable_hardware_rendering(void);

/*
 * @brief Disables hardware rendering
 */
void display_vglite_disable_hardware_rendering(void);

/*
 * @brief Checks if hardware rendering is enabled
 *
 * @return: true if hardware rendering is enabled, false otherwise
 */
bool display_vglite_is_hardware_rendering_enabled(void);

/*
 * @brief Toggles hardware rendering
 */
void display_vglite_toggle_hardware_rendering(void);

/*
 * @brief Create display, window contexes and initialize the VGLite library
 *
 * @return: a VG_LITE_XXX error in case of error, VG_LITE_SUCCESS otherwise
 */
vg_lite_error_t display_vglite_init(void);

/*
 * @brief Clean up VGLite library context
 */
void display_vglite_cleanup(void);

/*
 * @brief Get the main VGLite window
 */
vg_lite_window_t* display_vglite_get_window(void);

/*
 * @brief Get the current graphics buffer
 */
vg_lite_buffer_t *display_vglite_get_current_graphics_buffer(void);

/*
 * @brief Get the next graphics buffer to use
 */
vg_lite_buffer_t *display_vglite_get_next_graphics_buffer(void);

/*
 * Because of VGLite library implementation,
 * a buffer located at the same address should not be modified and reused
 * as the VGLite library check if the buffer address match with the one
 * stored in its context. If matching is successfull, render target will not
 * be updated.
 * This function rotate over an array of vg_lite_buffers to give a different
 * buffer address at each call and configures it from a MICROUI_Image.
 * The caller MUST send it to the vg_lite library to update its context.
 *
 * @param[in] gc: graphics context of destination
 *
 * @return a vg_lite buffer that should not match vg_lite library render context
 */
vg_lite_buffer_t* display_vglite_configure_destination(MICROUI_GraphicsContext* gc);

/*
 * This function configures a source buffer from a MICROUI_Image.
 *
 * @param[in] buffer: source buffer
 * @param[in] image: source image
 *
 * @return false if source image format is not supported by vg_lite, true on success
 */
bool display_vglite_configure_source(vg_lite_buffer_t *buffer, MICROUI_Image* image);

/*
 * @brief RT595 Porter-Duff operators seems to not be functional
 * When using transparency, the colors passed to the RT595
 * need to be modifed to display the correct colors.
 * This modification also has an impact on the front-panel
 * that needs to be fixed.
 *
 * Note: This has been tested on SRC_OVER only.
 *
 * @param[in] color: the color in ARGB8888 format to convert
 * @return the converted color int ARGB8888 format
 */
uint32_t display_vglite_porter_duff_workaround_ARGB8888(uint32_t color);

#endif // !defined DISPLAY_VGLITE_H

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

