/*
 * C
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include "fsl_dma.h"

#include "vglite_window.h"

#include "LLUI_DISPLAY.h"

#include "display_dma.h"

#include "interrupts.h"

#include "mej_log.h"

#include "power_manager.h"

#include "trace_platform.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/* @brief Maximum DMA transfers (DMA spec) */
#define DISPLAY_DMA_MAX_TRANSFER        (4096)

/* @brief Width of one element in a DMA transfer (DMA spec) */
#define DISPLAY_DMA_TRANSFER_WIDTH      (4U)

/* @brief Number of lines per DMA transfer */
#define DISPLAY_DMA_NB_LINES            (DISPLAY_DMA_MAX_TRANSFER / FRAME_BUFFER_STRIDE_BYTE)

/* @brief Number of descriptors in the DMA chain */
#define DISPLAY_DMA_NB_DESCS            \
	(1 + (FRAME_BUFFER_HEIGHT / DISPLAY_DMA_NB_LINES))

/* @brief Length of a DMA transfer */
#define DISPLAY_DMA_TRANSFER_LENGTH     \
	(DISPLAY_DMA_NB_LINES * (FRAME_BUFFER_WIDTH * FRAME_BUFFER_BYTE_PER_PIXEL))

/* @brief Number of DMA transfers in a chain */
#define DISPLAY_DMA_NB_TRANSFER         \
	(DISPLAY_DMA_TRANSFER_LENGTH / DISPLAY_DMA_TRANSFER_WIDTH)

/* @brief Number of lines of last DMA transfer */
#define DISPLAY_LAST_DMA_NB_LINES            (FRAME_BUFFER_HEIGHT % DISPLAY_DMA_NB_LINES)

/* @brief Length of a Last DMA transfer */
#define DISPLAY_LAST_DMA_TRANSFER_LENGTH     \
	(DISPLAY_LAST_DMA_NB_LINES * (FRAME_BUFFER_WIDTH * FRAME_BUFFER_BYTE_PER_PIXEL))

/* @brief Number of LastDMA transfers in a chain */
#define DISPLAY_LAST_DMA_NB_TRANSFER         \
	(DISPLAY_LAST_DMA_TRANSFER_LENGTH / DISPLAY_DMA_TRANSFER_WIDTH)

// -----------------------------------------------------------------------------
// Static Variables
// -----------------------------------------------------------------------------

#if DISPLAY_DMA_ENABLED != 0
static dma_handle_t g_DMA_Handle;

#if defined(__ICCARM__)
#pragma data_alignment = 16U
static dma_descriptor_t dma_descriptors[2 * DISPLAY_DMA_NB_DESCS];
#elif defined(__CC_ARM)
static dma_descriptor_t __attribute__((aligned(16U))) dma_descriptors[2 * DISPLAY_DMA_NB_DESCS];
#elif defined(__GNUC__)
static dma_descriptor_t __attribute__((aligned(16U))) dma_descriptors[2 * DISPLAY_DMA_NB_DESCS];
#endif

#endif // DISPLAY_DMA_ENABLED != 0

// -----------------------------------------------------------------------------
// Internal function definitions
// -----------------------------------------------------------------------------

/*
 * @brief: Check buffer is known
 *
 * @param b: Buffer to check
 */
static int __dma_check_buffer(framebuffer_t *b);

/*
 * @brief: Callback called when a DMA transfer is done
 *
 * @param[in] handle: Handle on the DMA transfer
 * @param[in] param: User parameters for the callback
 * @param[in] transfer_done: Flag indicating if the current transfer is done
 */
static void __dma_callback(dma_handle_t *handle, void *param, bool transfer_done, uint32_t tcds);

// -----------------------------------------------------------------------------
// Public functions
// -----------------------------------------------------------------------------

#if DISPLAY_DMA_ENABLED != 0
// See the header file for the function documentation
void display_dma_initialize(const framebuffer_t * framebuffers[]) {

	// MEJ_DEBUG_enable(MEJ_DEBUG_DISPLAY_DMA_LVL);

	DMA_Init(DMA1);
	DMA_CreateHandle(&g_DMA_Handle, DMA1, 0);
	DMA_EnableChannel(DMA1, 0);
	DMA_SetCallback(&g_DMA_Handle, __dma_callback, NULL);

	// Initialize the dma descriptors
	for (int i = 0; i < 2; i++) {
		framebuffer_t const * fb_src = framebuffers[i];
		framebuffer_t const * fb_dst = framebuffers[1-i];
		int offset = i*DISPLAY_DMA_NB_DESCS;
		int y, d;

		d = offset;

		do {
			y = DISPLAY_DMA_NB_LINES * (d - offset);

			if (d >= (offset+(DISPLAY_DMA_NB_DESCS-1))) {
				break;
			}

			DMA_SetupDescriptor(
					&(dma_descriptors[d]),
					DMA_CHANNEL_XFER(
						true, false, false, false,
						DISPLAY_DMA_TRANSFER_WIDTH,
						kDMA_AddressInterleave1xWidth,
						kDMA_AddressInterleave1xWidth,
						DISPLAY_DMA_TRANSFER_LENGTH
						),
					(void *) &(fb_src->p[y]),
					(void *) &(fb_dst->p[y]),
					&(dma_descriptors[d+1])
					);

			d++;
		} while (1);

		// Check if we need a last descriptor to send last lines
#if (0 != DISPLAY_LAST_DMA_NB_TRANSFER)
			y = DISPLAY_DMA_NB_LINES * (d - offset);
			// Setup last descriptor
			DMA_SetupDescriptor(
					&(dma_descriptors[d]),
					DMA_CHANNEL_XFER(
						false, false, true, false,
						DISPLAY_DMA_TRANSFER_WIDTH,
						kDMA_AddressInterleave1xWidth,
						kDMA_AddressInterleave1xWidth,
						DISPLAY_LAST_DMA_NB_TRANSFER
						),
					(void *) &(fb_src->p[y]),
					(void *) &(fb_dst->p[y]),
					NULL);
#else
			// Reconfigure previous descriptor to be the last descriptor
			DMA_SetupDescriptor(
					&(dma_descriptors[d-1]),
					DMA_CHANNEL_XFER(
						false, false, true, false,
						DISPLAY_DMA_TRANSFER_WIDTH,
						kDMA_AddressInterleave1xWidth,
						kDMA_AddressInterleave1xWidth,
						DISPLAY_DMA_TRANSFER_LENGTH
						),
					(void *) &(fb_src->p[y]),
					(void *) &(fb_dst->p[y]),
					NULL);
#endif
	}
}

// See the header file for the function documentation
void display_dma_start(framebuffer_t *src, framebuffer_t *dst, int ymin, int ymax) {
	dma_descriptor_t    *p_dma_desc;

	power_manager_enable_low_power(POWER_DMA, LOW_POWER_FORBIDDEN);

	TRACE_HW_TASK_START(HW_TASK_DMA_ID);

	__dma_check_buffer(src);
	__dma_check_buffer(dst);

	if (src == s_frameBufferAddress[1]) {
		p_dma_desc = &dma_descriptors[DISPLAY_DMA_NB_DESCS];
	} else {
		p_dma_desc = &dma_descriptors[0];
	}

	DMA_SubmitChannelDescriptor(&g_DMA_Handle, p_dma_desc);
	DMA_StartTransfer(&g_DMA_Handle);
}

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

// See the section 'Internal function definitions' for the function documentation
static void __dma_callback(dma_handle_t *handle, void *param, bool transfer_done, uint32_t tcds)
{
	if (transfer_done)
	{
		uint8_t it = interrupt_enter();
		LLUI_DISPLAY_flushDone(true);
		power_manager_enable_low_power(POWER_DMA, LOW_POWER_AUTHORIZED);
                
        // Request power_manager to power down this buffer
		power_manager_framebuffer_request_power_off();

		TRACE_HW_TASK_STOP(HW_TASK_DMA_ID);

		interrupt_leave(it);
	}
}

// See the section 'Internal function definitions' for the function documentation
static int __dma_check_buffer(framebuffer_t *b) {
	int found_fb = 0;

	for (   int i = 0;
			i < (sizeof(s_frameBufferAddress) / sizeof(s_frameBufferAddress[0]));
			i++)
	{
		if (b == s_frameBufferAddress[i]) {
			found_fb = 1;
		}
	}

	if (0 == found_fb) {
		MEJ_LOG_ERROR("dma address error: 0x%p unkown\n", b);
		return -1;
	}

	return 0;
}

#endif // DISPLAY_DMA_ENABLED != 0

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

