/*
 * C
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include <stdint.h>

#include "vg_lite_hw.h"

#include "LLUI_DISPLAY.h"
#include "LLUI_DISPLAY_impl.h"
#include "LLUI_PAINTER_impl.h"

#include "display_vglite.h"
#include "display_framebuffer.h"

#include "color.h"

#include "event_generator.h"

#include "vglite_support.h"
#include "vglite_window.h"

#include "mej_log.h"

#include "power_manager.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/*
 * @brief To check if the image format is known
 */
#define VG_LITE_UNKNOWN_FORMAT		((vg_lite_buffer_format_t) -1)

/*
 * @brief Width of the Tesselation window
 *
 * @Warning: this impacts the VGLite allocation size
 */
#define DISPLAY_VGLITE_TESSELATION_WIDTH	256

/*
 * @brief Height of the Tesselation window
 *
 * @Warning: this impacts the VGLite allocation size
 */
#define DISPLAY_VGLITE_TESSELATION_HEIGHT	256

// -----------------------------------------------------------------------------
// Static Variables
// -----------------------------------------------------------------------------

/*
 * @brief VGLite buffer array cache
 */
static vg_lite_buffer_t buffer_array[3]; // {dest,dest,src}

/*
 * @brief Index in the VGLite buffer array cache
 */
static uint32_t buffer_index;

/*
 * @brief VGLite display context
 */
static vg_lite_display_t display;

/*
 * @brief VGLite window context
 */
static vg_lite_window_t window;

/*
 * @brief flag to enable/disable hardware rendering
 */
static bool hardware_rendering;

/*
 * @brief flag to enable/disable hardware rendering
 */
static bool vg_lite_operation_started;

/*
 * @brief vglite operation semaphore
 */
SemaphoreHandle_t display_vglite_operation;

/*
 * @brief current MicroUI Graphics Context where render the drawing
 */
static MICROUI_GraphicsContext* current_gc;

// -----------------------------------------------------------------------------
// Static Constants
// -----------------------------------------------------------------------------

/*
 * @brief LUT to convert MicroUI image format to VGLite image format
 */
const vg_lite_buffer_format_t __microui_to_vg_lite_format[] = {
	VG_LITE_RGB565,	        // MICROUI_IMAGE_FORMAT_LCD = 0,
	VG_LITE_UNKNOWN_FORMAT,	// UNKNOWN = 1,
	VG_LITE_RGBA8888,		// MICROUI_IMAGE_FORMAT_ARGB8888 = 2,
	VG_LITE_UNKNOWN_FORMAT,	// MICROUI_IMAGE_FORMAT_RGB888 = 3, unsupported
	VG_LITE_RGB565,			// MICROUI_IMAGE_FORMAT_RGB565 = 4,
	VG_LITE_RGBA5551,	// MICROUI_IMAGE_FORMAT_ARGB1555 = 5
	VG_LITE_RGBA4444,		// MICROUI_IMAGE_FORMAT_ARGB4444 = 6,
	VG_LITE_A4,				// MICROUI_IMAGE_FORMAT_A4 = 7,
	VG_LITE_A8,				// MICROUI_IMAGE_FORMAT_A8 = 8,
	// outside of the table ... MICROUI_IMAGE_FORMAT_CUSTOM = 255,
};

// -----------------------------------------------------------------------------
// Internal function definitions
// -----------------------------------------------------------------------------

/*
 * @brief This callback must be called when the GCNanoLite-V gpu irq is
 * triggered
 *
 * @param[in] flags: the gpu irq flags
 */
static void __gpu_irq_callback(uint32_t flags);

/*
 * @brief Configures a vg_lite buffer for a RGB565 image
 *
 * @param[in] b: buffer to configure
 */
static void __buffer_default_configuration2(vg_lite_buffer_t* b);

/*
 * @brief Configures a vg_lite buffer for a RGB565 image
 *
 * @param[in] buffer_index: index of the buffer to configure
 */
static void __buffer_default_configuration(uint32_t buffer_index);

/*
 * @brief Configures and return current vg_lite buffer for the given image (address and size)
 *
 * @param[in] image: image from which buffer should be configured
 * @param[in] b: buffer to configure
 */
static void __buffer_set_address_and_size2(MICROUI_Image* image, vg_lite_buffer_t* b);

/*
 * @brief Configures and return current vg_lite buffer for the given image (address and size)
 *
 * @param[in] image: MicroUI image from which buffer should be configured
 * @param[in] i: index of the buffer to configure
 */
static vg_lite_buffer_t* __buffer_set_address_and_size(MICROUI_Image* image, uint32_t buffer_index);

/*
 * @brief Convert MICROUI_ImageFormat to vg_lite_buffer_format_t
 *
 * @param[in] microui_format: MicroUI image format
 */
static vg_lite_buffer_format_t __convert_format(MICROUI_ImageFormat microui_format);

/*
 * @brief Special function that manage overlay and rendering mode
 *
 * @param[in] params: contain which button is pressed when triggering the special function
 */
static void __vglite_special_function(void *params);

// -----------------------------------------------------------------------------
// Public functions
// -----------------------------------------------------------------------------

// See the header file for the function documentation
void display_vglite_cleanup(void) {
	vg_lite_close();
}

// See the header file for the function documentation
vg_lite_error_t display_vglite_init(void) {
	vg_lite_error_t ret = VG_LITE_SUCCESS;

	ret = VGLITE_CreateDisplay(&display);
	if (VG_LITE_SUCCESS != ret) {
		MEJ_LOG_ERROR("VGLITE_CreateDisplay failed: VGLITE_CreateDisplay() returned error %d", ret);
		return ret;
	}

	// Initialize the window.
	ret = VGLITE_CreateWindow(&display, &window);
	if (VG_LITE_SUCCESS != ret) {
		MEJ_LOG_ERROR("VGLITE_CreateWindow failed: VGLITE_CreateWindow() returned error %d", ret);
		return ret;
	}

	// Initialize the VGLite library and it's tesselation buffer.
	ret = vg_lite_init(DISPLAY_VGLITE_TESSELATION_WIDTH, DISPLAY_VGLITE_TESSELATION_HEIGHT);
	if (VG_LITE_SUCCESS != ret) {
		MEJ_LOG_ERROR("vg_lite engine init failed: vg_lite_init() returned error %d", ret);
		display_vglite_cleanup();
		return ret;
	}

	// Initialize the VGLite buffer cache
	for (int i = 0; i < sizeof(buffer_array) / sizeof(buffer_array[0]); i++) {
		__buffer_default_configuration(i);
	}

	buffer_index = 0;

	// Enable by default hardware rendering
	display_vglite_enable_hardware_rendering();

	// Register callback to event generator to enable / disable hardware rendering
	EVENT_GENERATOR_register_special_cb(__vglite_special_function, 0);

	display_vglite_operation = xSemaphoreCreateBinary();

	vg_lite_hal_register_irq_callback(&__gpu_irq_callback);

	current_gc = NULL;

	return VG_LITE_SUCCESS;
}

// See the header file for the function documentation
vg_lite_window_t* display_vglite_get_window(void) {
	return &window;
}

// See the header file for the function documentation
vg_lite_buffer_t *display_vglite_get_current_graphics_buffer(void) {
	return VGLITE_GetRenderTarget(&window);
}

// See the header file for the function documentation
vg_lite_buffer_t *display_vglite_get_next_graphics_buffer(void) {
	return VGLITE_GetNextBuffer(&window);
}

void display_vglite_enable_hardware_rendering(void)
{
	MEJ_LOG_INFO("Hardware rendering enabled\n");
	hardware_rendering = true;
}

// See the header file for the function documentation
void display_vglite_disable_hardware_rendering(void)
{
	MEJ_LOG_INFO("Hardware rendering disabled\n");
	hardware_rendering = false;
}

// See the header file for the function documentation
void display_vglite_toggle_hardware_rendering(void)
{
	hardware_rendering = !hardware_rendering;

	MEJ_LOG_INFO("Hardware rendering toggle: %d\n", hardware_rendering);
}

// See the header file for the function documentation
bool display_vglite_is_hardware_rendering_enabled(void)
{
	return hardware_rendering;
}

// See the header file for the function documentation
vg_lite_buffer_t* display_vglite_configure_destination(MICROUI_GraphicsContext* gc)
{
	MICROUI_Image* image = &gc->image;
	current_gc = LLUI_DISPLAY_isLCD(&gc->image) ? gc : NULL;;

	if ((uint32_t)LLUI_DISPLAY_getBufferAddress(image) != buffer_array[buffer_index].address)
	{
		// we target another destination than previous drawing
		// -> have to use another context to force vg_lite reloading the new context

		buffer_index++;
		buffer_index &= 1; // 0 & 1 only

		vg_lite_buffer_t* blit_context = __buffer_set_address_and_size(image, buffer_index);

		return blit_context;
	}
	else
	{
		//we target the same destination than previous drawing: nothing to do
		return &buffer_array[buffer_index];
	}
}

// See the header file for the function documentation
bool display_vglite_configure_source(vg_lite_buffer_t *buffer, MICROUI_Image* image)
{

#if FRAME_BUFFER_LINE_ALIGN_BYTE < VGLITE_IMAGE_LINE_ALIGN_BYTE
	if (LLUI_DISPLAY_isLCD(image)){
		// frame buffer does not respect VGLite alignment
		return false;    
	}
#endif

	vg_lite_buffer_format_t format = __convert_format((MICROUI_ImageFormat)image->format);

	if (VG_LITE_UNKNOWN_FORMAT == format) {
		return false;
	}

	__buffer_default_configuration2(buffer);
	__buffer_set_address_and_size2(image, buffer);

	buffer->image_mode = VG_LITE_MULTIPLY_IMAGE_MODE; // image only
	buffer->transparency_mode = VG_LITE_IMAGE_TRANSPARENT; // FIXME
	buffer->format = format;
	buffer->stride = LLUI_DISPLAY_getStrideInBytes(image);

	return true;
}

// See the header file for the function documentation
void display_vglite_start_operation(void) {
	power_manager_enable_low_power(POWER_GPU, LOW_POWER_FORBIDDEN);
	vg_lite_operation_started = true;
}

// See the header file for the function documentation
void display_vglite_cancel_operation(void) {
	vg_lite_operation_started = false;
	power_manager_enable_low_power(POWER_GPU, LOW_POWER_AUTHORIZED);
}

// See the header file for the function documentation
void display_vglite_wait_operation(void) {
	xSemaphoreTake(display_vglite_operation, portMAX_DELAY);
	power_manager_enable_low_power(POWER_GPU, LOW_POWER_AUTHORIZED);
}

// See the header file for the function documentation
uint32_t display_vglite_porter_duff_workaround_ARGB8888(uint32_t color) {
	uint8_t alpha = COLOR_GET_CHANNEL(color, ARGB8888, ALPHA);

	if (alpha == 0) {
		return 0;
	}

	uint8_t red = (
			(alpha * COLOR_GET_CHANNEL(color, ARGB8888, RED))
			>> COLOR_ARGB8888_CHANNEL_SIZE);
	uint8_t green = (
			(alpha * COLOR_GET_CHANNEL(color, ARGB8888, GREEN))
			>> COLOR_ARGB8888_CHANNEL_SIZE);
	uint8_t blue = (
			(alpha * COLOR_GET_CHANNEL(color, ARGB8888, BLUE))
			>> COLOR_ARGB8888_CHANNEL_SIZE);

	return COLOR_SET_COLOR(alpha, red, green, blue, ARGB8888);
}

/*
 * Updates current Graphics Context with vglite rendering area
 */
void vg_lite_draw_notify_render_area(uint32_t x, uint32_t y, uint32_t right, uint32_t bottom)
{
	MICROUI_GraphicsContext* gc = current_gc;
	if (NULL != gc)
	{
		LLUI_DISPLAY_setDrawingLimits(gc, x, y, right, bottom);
	}
}

// -----------------------------------------------------------------------------
// Java APIs
// -----------------------------------------------------------------------------

/*
 * @brief Checks if hardware rendering is enabled
 *
 * @return true if hardware rendering is enabled, false otherwise
 */
jboolean Java_com_microej_display_utils_NHardwareRendering_isEnabled(void) {
	return hardware_rendering == true;
}

/*
 * @brief Checks if hardware rendering is disabled
 *
 * @return true if hardware rendering is disabled, false otherwise
 */
jboolean Java_com_microej_display_utils_NHardwareRendering_isDisabled(void) {
	return hardware_rendering == false;
}

/*
 * @brief Disables hardware rendering
 */
void Java_com_microej_display_utils_NHardwareRendering_disable(void) {
	display_vglite_disable_hardware_rendering();
}

/*
 * @brief Enables hardware rendering
 */
void Java_com_microej_display_utils_NHardwareRendering_enable(void) {
	display_vglite_enable_hardware_rendering();
}

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

// See the section 'Internal function definitions' for the function documentation
static void __gpu_irq_callback(uint32_t flags) {
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

	if (vg_lite_operation_started) {
		vg_lite_operation_started = false;
		//LLUI_DISPLAY_notifyAsynchronousDrawingEnd(true);

		xSemaphoreGiveFromISR(display_vglite_operation, &xHigherPriorityTaskWoken);
		if(xHigherPriorityTaskWoken != pdFALSE )
		{
			// Force a context switch here.
			portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
		}
	}
}

// See the section 'Internal function definitions' for the function documentation
static void __buffer_default_configuration2(vg_lite_buffer_t* b)
{
      /* Reset planar. */
        b->yuv.uv_planar =
        b->yuv.v_planar =
        b->yuv.alpha_planar = 0;
  
	b->tiled = VG_LITE_LINEAR;
	b->format = VG_LITE_RGB565;
	b->handle = NULL;
	b->memory = NULL;
	b->address = 0;
	b->yuv.swizzle = VG_LITE_SWIZZLE_UV;
	b->image_mode = VG_LITE_NORMAL_IMAGE_MODE;
	b->transparency_mode = VG_LITE_IMAGE_OPAQUE;
}

// See the section 'Internal function definitions' for the function documentation
static void __buffer_default_configuration(uint32_t i)
{
	vg_lite_buffer_t* b = &buffer_array[i];

	__buffer_default_configuration2(b);
}

// See the section 'Internal function definitions' for the function documentation
static void __buffer_set_address_and_size2(MICROUI_Image* image, vg_lite_buffer_t* b)
{
	b->width = image->width;
	b->height = image->height;
	b->stride = LLUI_DISPLAY_getStrideInBytes(image);
	b->memory = (void*)LLUI_DISPLAY_getBufferAddress(image);
	b->address = (uint32_t)LLUI_DISPLAY_getBufferAddress(image);
}

// See the section 'Internal function definitions' for the function documentation
static vg_lite_buffer_t* __buffer_set_address_and_size(MICROUI_Image* image, uint32_t i)
{
	vg_lite_buffer_t* b = &buffer_array[i];

	__buffer_set_address_and_size2(image, b);
	return b;
}

// See the section 'Internal function definitions' for the function documentation
vg_lite_buffer_format_t __convert_format(MICROUI_ImageFormat microui_format)
{
	vg_lite_buffer_format_t vg_lite_format = VG_LITE_UNKNOWN_FORMAT;
	if (microui_format < (sizeof(__microui_to_vg_lite_format) / sizeof(__microui_to_vg_lite_format[0]))) {
		vg_lite_format = __microui_to_vg_lite_format[microui_format];
	}

	return vg_lite_format;
}

// See the section 'Internal function definitions' for the function documentation
static void __vglite_special_function(void *params) {
	display_vglite_toggle_hardware_rendering();
}

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

