/*
 * Java
 *
 * Copyright 2021 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.demo.watch.util.path;

import com.microej.vivante.utils.PathBufferFP32;
import com.verisilicon.vivante.vg_lite;
import com.verisilicon.vivante.vg_lite_matrix;
import com.verisilicon.vivante.vg_lite_path;

import ej.microui.display.GraphicsContext;

/**
 * A vector graphics that represents a filled circle.
 */
public class FilledCircle implements AutoCloseable {

	private static final float TANGENT = 0.5522847498307933F; // Magic number to approximate circle arcs from Bezier

	private static final vg_lite_path PATH = computePath();

	private final vg_lite_matrix matrix;
	private final float radius;

	/**
	 * Creates a filled circle.
	 *
	 * @param radius
	 *            the radius of the circle.
	 */
	public FilledCircle(float radius) {
		this.matrix = new vg_lite_matrix();
		this.radius = radius;
	}

	@Override
	public void close() {
		this.matrix.close();
	}

	/**
	 * Renders this filled circle.
	 *
	 * @param g
	 *            the graphics context to use.
	 * @param x
	 *            the x coordinate of the center of the circle.
	 * @param y
	 *            the y coordinate of the center of the circle.
	 */
	public void render(GraphicsContext g, float x, float y) {
		// Create matrix
		vg_lite_matrix matrix = this.matrix;
		vg_lite_matrix.vg_lite_identity(matrix);
		vg_lite_matrix.vg_lite_translate(g.getTranslationX() + x, g.getTranslationY() + y, matrix);
		vg_lite_matrix.vg_lite_scale(this.radius, this.radius, matrix);

		// Draw path
		vg_lite.vg_lite_draw_path(g, PATH, vg_lite.VG_LITE_FILL_EVEN_ODD, matrix, vg_lite.VG_LITE_BLEND_NONE,
				0xff000000 | g.getColor());
	}

	private static vg_lite_path computePath() {
		// Create path: approximation of a circle using Bezier cubic curves
		PathBufferFP32 pb = new PathBufferFP32();
		pb.move(0, -1.0f);
		pb.cubicRel(TANGENT, 0, 1.0f, 1.0f - TANGENT, 1.0f, 1.0f);
		pb.cubicRel(0, TANGENT, -1.0f + TANGENT, 1.0f, -1.0f, 1.0f);
		pb.cubicRel(-TANGENT, 0, -1.0f, -1.0f + TANGENT, -1.0f, -1.0f);
		pb.cubicRel(0, -TANGENT, 1.0f - TANGENT, -1.0f, 1.0f, -1.0f);
		return pb.toVGLitePath(-1.0f, -1.0f, 1.0f, 1.0f, vg_lite_path.VG_LITE_UPPER);
	}
}
