/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor;

import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ElementVisitor;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.PackageElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.TypeKind;
import javax.lang.model.type.TypeMirror;

import com.is2t.giml.annotation.processor.model.Attribute;
import com.is2t.giml.annotation.processor.model.Child;
import com.is2t.giml.annotation.processor.model.CodeInterface;
import com.is2t.giml.annotation.processor.model.CodeInterface.CodeMethodPrototype;
import com.is2t.giml.annotation.processor.model.Description;

import ej.giml.annotation.ElementAttribute;
import ej.giml.annotation.ElementChild;
import ej.giml.annotation.ElementConstructor;
import ej.giml.annotation.ElementRequiredAttribute;
import ej.giml.annotation.ElementSetter;

/**
 *
 */
public class GIMLElementVisitor implements ElementVisitor<Void, Boolean> {

	private static final String REFERENCE_TYPE = "reference";
	private static final String CODE_TYPE = "code";
	private static final String QUOTE_IN_STRING = "\"";
	private static final String ESCAPED_QUOTE_IN_STRING = "\\\"";

	private static final Logger LOGGER = Logger.getLogger(GIMLElementVisitor.class.getName());

	private Description _description; // NOSONAR
	private boolean _constructorVisited = false; // NOSONAR

	public Description getDescription() {
		if (this._description != null && this._description.getName() != null) {
			return this._description;
		}
		return null;
	}

	@Override
	public Void visit(Element e, Boolean p) {
		// nothing to do
		return null;
	}

	@Override
	public Void visit(Element e) {
		// nothing to do
		return null;
	}

	@Override
	public Void visitPackage(PackageElement e, Boolean p) {
		// nothing to do
		return null;
	}

	@Override
	public Void visitType(TypeElement e, Boolean p) {
		ej.giml.annotation.Element classAnnot = e.getAnnotation(ej.giml.annotation.Element.class);
		if (classAnnot != null && p.booleanValue()) {
			// Type is annotated. Visit all enclosing elements (methods, fields, ...) to
			// find GIML attributes,
			// containers,...
			this.initDescription();

			String elName = this.getElementName(classAnnot.name(), e);
			boolean isRoot = classAnnot.root();
			String elFriendlyName = this.getValueOrDefault(classAnnot.friendlyName(), elName);
			String elDescription = this.getValueOrDefault(classAnnot.description(), elName);
			String elClassName = e.getQualifiedName().toString();
			this._description.setName(elName);
			this._description.setRoot(isRoot);
			this._description.setFriendlyName(elFriendlyName);
			this._description.setDescription(elDescription);
			this._description.setElementClassName(elClassName);
		}

		List<? extends Element> enclosedElements = e.getEnclosedElements();
		for (Element element : enclosedElements) {
			element.accept(this, p);
		}

		TypeMirror superclass = e.getSuperclass();
		if (superclass.getKind() == TypeKind.DECLARED) {
			DeclaredType superType = (DeclaredType) superclass;
			superType.asElement().accept(this, Boolean.FALSE);
		}
		return null;
	}

	@Override
	public Void visitVariable(VariableElement e, Boolean p) {
		if (e.getKind() == ElementKind.PARAMETER) {
			// Handle attributes with default values
			ElementAttribute attributeAnnot = e.getAnnotation(ElementAttribute.class);
			if (attributeAnnot != null) {
				this.initDescription();

				String attrName = this.getElementName(attributeAnnot.name(), e);
				String type = (attributeAnnot.reference() ? REFERENCE_TYPE : e.asType().toString());
				String defaultValue = attributeAnnot.defaultValue();
				String normalizedDefaultValue = normalizeDefaultValue(attributeAnnot.defaultValue(), type);
				Attribute attribute = new Attribute(attrName, defaultValue, normalizedDefaultValue, type, true);
				addAttribute(this._description, attribute);
			}
			// Handle required attributes
			ElementRequiredAttribute requiredAttributeAnnot = e.getAnnotation(ElementRequiredAttribute.class);
			if (requiredAttributeAnnot != null) {
				this.initDescription();

				String attrName = this.getElementName(requiredAttributeAnnot.name(), e);
				String type = (requiredAttributeAnnot.reference() ? REFERENCE_TYPE : e.asType().toString());
				Attribute attribute = new Attribute(attrName, type, true);
				addAttribute(this._description, attribute);
			}
		}
		return null;
	}

	@Override
	public Void visitExecutable(ExecutableElement e, Boolean p) {
		if (e.getKind() == ElementKind.CONSTRUCTOR && p.booleanValue()) {
			ElementConstructor constructorAnnot = e.getAnnotation(ElementConstructor.class);
			if (constructorAnnot != null) {
				if (this._constructorVisited) {
					throw new IllegalArgumentException(String.format("Cannot declare multiple '%s' in element '%s'",
							ElementConstructor.class.getSimpleName(), this._description.getElementClassName()));
				}
				// Visit all constructor parameters to find GIML attributes
				List<? extends VariableElement> parameters = e.getParameters();
				for (VariableElement parameter : parameters) {
					parameter.accept(this, null);
				}
				this._constructorVisited = true;
			}
		}

		if (e.getKind() == ElementKind.METHOD) {
			ElementChild childAnnot = e.getAnnotation(ElementChild.class);
			if (childAnnot != null) {
				// Method is annotated with Child annotation; so this is a GIML container
				// declaration
				this.initDescription();

				String childName = childAnnot.name();
				String foundPrefix = null;
				if (childName == null || childName.trim().isEmpty()) {
					String[] prefixes = { childAnnot.prefix(), "set", "add" };
					String[] prefixAndName = extractElementNameWithPrefix(e, prefixes);
					foundPrefix = prefixAndName[0];
					childName = prefixAndName[1].toLowerCase();
				}

				List<? extends VariableElement> parameters = e.getParameters();
				Deque<Attribute> params = new LinkedList<Attribute>();
				for (int i = 0; i < parameters.size(); i++) {
					VariableElement param = parameters.get(i);
					params.addLast(new Attribute(param.getSimpleName().toString(),
							param.asType().accept(new GIMLTypeVisitor(), null), false));
				}
				Child child = new Child(childName, foundPrefix, childAnnot.cardinality(), childAnnot.ordered(),
						childAnnot.disabled(), params);
				this._description.addChild(child);

			} else if (_description != null) {
				// visiting an element: maybe add code attributes
				loadCodeAttributes(e);
			}

			ElementSetter setterAnnot = e.getAnnotation(ElementSetter.class);
			if (setterAnnot != null) {
				// Not supported (yet)
				LOGGER.log(Level.WARNING, "'ElementSetter' annotation not supported. Attribute will not be created.");
			}
		}
		return null;
	}

	/**
	 * Load optional "hook/listener" code attributes, with code content written in
	 * javascript. Pattern: a type that declares a method '[add|set]AType(AType)',
	 * where 'AType' is an interface. All methods declared by the interface are
	 * added as non required attributes.
	 */
	private void loadCodeAttributes(ExecutableElement e) {
		CodeInterface codeInterface = getAssociatedCodeInterface(e);
		if (codeInterface == null) {
			return;
		}

		// register this code interface to the type description
		_description.addCodeInterface(codeInterface);

		// create all new attributes related to this code interface
		addCodeAttributes(codeInterface);
	}

	/**
	 * Add to the current {@link Description} all new attributes related to the
	 * given {@link CodeInterface}
	 *
	 * @param codeInterface
	 *            a code interface detected for this {@link Description}
	 */
	private void addCodeAttributes(CodeInterface ci) {
		for (CodeMethodPrototype m : ci.getMethods()) {
			addAttribute(_description, new Attribute(m.getName(), "", "", CODE_TYPE, false));
		}
	}

	private void addAttribute(Description description, Attribute attribute) {
		description.addAttribute(attribute);
		if (!attribute.getAttributeClassName().equals(CODE_TYPE)) {
			// add its derived code attribute definition (it is not a code attribute)
			description.addAttribute(new Attribute(attribute.getName() + "Code", "", "", CODE_TYPE, false));
		}
	}

	/**
	 *
	 * @param codeInterface
	 * @param methodsVect
	 * @return false if one of the methods does not match the pattern
	 */
	private boolean buildCodeInterface(TypeElement codeInterface, ArrayList<CodeMethodPrototype> methodsVect) {
		// add attributtes for direct methods
		List<? extends Element> interfaceElements = codeInterface.getEnclosedElements();
		for (Element interfaceElement : interfaceElements) {
			if (interfaceElement.getKind() == ElementKind.METHOD
					&& !interfaceElement.getModifiers().contains(Modifier.STATIC)) {
				// this is an instance method of the code interface
				// => the simple name is a valid attribute
				// check void / void method
				ExecutableElement methodElement = (ExecutableElement) interfaceElement;
				if (methodElement.getParameters().size() > 0) {
					return false;
				}
				if (methodElement.getReturnType().getKind() != TypeKind.VOID) {
					return false;
				}
				methodsVect.add(new CodeMethodPrototype(interfaceElement.getSimpleName().toString()));
			}
		}
		// process super interfaces
		for (TypeMirror superInterface : codeInterface.getInterfaces()) {
			if (!buildCodeInterface((TypeElement) ((DeclaredType) superInterface).asElement(), methodsVect)) {
				return false;
			}
		}

		return true;
	}

	/**
	 * From an element method, gets the associated {@link CodeInterface} if it
	 * exists.
	 *
	 * @param e
	 *            the method element
	 * @return null if the method does not match the code interface pattern
	 */
	private CodeInterface getAssociatedCodeInterface(ExecutableElement e) {
		String[] prefixes = { "set", "add" };
		String[] prefixAndName = extractElementNameWithPrefix(e, prefixes);
		// check add/set method name convention
		if (prefixAndName == null) {
			return null;
		}
		String basename = prefixAndName[1];
		List<? extends VariableElement> parameters = e.getParameters();

		// check single argument
		if (parameters.size() != 1) {
			return null;
		}
		VariableElement param = parameters.get(0);

		// check argument type is an interface that matches the basename
		TypeMirror t = param.asType();
		if (t.getKind() != TypeKind.DECLARED) {
			return null;
		}
		TypeElement typeElement = (TypeElement) ((DeclaredType) t).asElement();
		if (!typeElement.getSimpleName().toString().equals(basename)) {
			return null;
		}
		if (!typeElement.getKind().isInterface()) {
			return null;
		}

		ArrayList<CodeMethodPrototype> methodsVect = new ArrayList<>();

		// check and build methods
		if (!buildCodeInterface(typeElement, methodsVect)) {
			return null;
		}
		return new CodeInterface(typeElement.getQualifiedName().toString(),
				methodsVect.toArray(new CodeMethodPrototype[methodsVect.size()]));
	}

	/**
	 * Extract the element name after removing one of the provided prefixes.
	 *
	 * @param e
	 *            the element
	 * @return an array of length 2 with the prefix and the name without the prefix,
	 *         or null if the element name does not starts with one the provided
	 *         prefixes
	 */
	private String[] extractElementNameWithPrefix(ExecutableElement e, String[] prefixes) {
		String methodName = e.getSimpleName().toString();
		for (String prefix : prefixes) {
			if (prefix != null && !prefix.isEmpty() && methodName.startsWith(prefix)) {
				methodName = methodName.substring(prefix.length());
				return new String[] { prefix, methodName };
			}
		}
		return null;
	}

	@Override
	public Void visitTypeParameter(TypeParameterElement e, Boolean p) {
		// nothing to do
		return null;
	}

	@Override
	public Void visitUnknown(Element e, Boolean p) {
		// nothing to do
		return null;
	}

	private void initDescription() {
		if (this._description == null) {
			this._description = new Description();
		}
	}

	private String getElementName(String elementName, Element element) {
		return getValueOrDefault(elementName, element.getSimpleName().toString());
	}

	private String getValueOrDefault(String value, String defaultValue) {
		if (value == null || value.trim().isEmpty()) {
			value = defaultValue;
		}
		return value;
	}

	private String normalizeDefaultValue(String defaultValue, String defaultValueType) {
		if ((String.class.getCanonicalName().equals(defaultValueType) || REFERENCE_TYPE.equals(defaultValueType))
				&& !"null".equalsIgnoreCase(defaultValue)) {
			// TODO Manage quotes in values.
			defaultValue = ESCAPED_QUOTE_IN_STRING + defaultValue + ESCAPED_QUOTE_IN_STRING;
		}
		return defaultValue;
	}

}
