/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor.generator;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;
import org.stringtemplate.v4.StringRenderer;

import com.is2t.giml.annotation.processor.model.Child;
import com.is2t.giml.annotation.processor.model.Description;
import com.is2t.giml.description.DescriptionPackage;

/**
 * The template helper class allows to manipulate description-related templates, and instantiate them with given values.
 */
public class TemplateHelper {

	/*
	 * Template files
	 */
	private static final String CLASSDESCRIPTION_TEMPLATE_PATH = "templates/classDescription.stg";
	private static final String CLASSDESCRIPTIONPACKAGE_TEMPLATE_PATH = "templates/classDescriptionPackage.stg";
	private static final String SERVICEDESCRIPTIONPACKAGE_TEMPLATE_PATH = "templates/serviceDescriptionPackage.stg";

	/*
	 * Template name
	 */
	private static final String TEMPLATE_CLASSDESCRIPTION = "ClassDescription";
	private static final String TEMPLATE_CLASSDESCRIPTIONPACKAGE = "ClassDescriptionPackage";
	private static final String TEMPLATE_SERVICEDESCRIPTIONPACKAGE = "ServiceDescriptionPackage";

	/*
	 * Template parameters
	 */
	private static final String TEMPLATE_PARAMETER_CHILDMETHODNAMES = "childMethodNames";
	private static final String TEMPLATE_PARAMETER_CLASSES = "classes";
	private static final String TEMPLATE_PARAMETER_CLASSNAME = "className";
	private static final String TEMPLATE_PARAMETER_DESCRIPTION = "description";
	private static final String TEMPLATE_PARAMETER_DESCRIPTIONS = "descriptions";
	private static final String TEMPLATE_PARAMETER_PACKAGE = "package";
	private static final String TEMPLATE_PARAMETER_TYPETOGIMLMAPPING = "typeToGimlMapping";
	private static final String TEMPLATE_PARAMETER_TYPETOJAVAMAPPING = "typeToJavaMapping";

	/*
	 * Other constants
	 */
	private static final String DEFAULT_ELEMENT = "element";
	private static final String DOUBLE_BREAK_LINE = "\n\n";

	private final STGroup _classDescription; // NOSONAR
	private final STGroup _classDescriptionPackage; // NOSONAR
	private final STGroup _serviceDescriptionPackage; // NOSONAR
	private final Map<String, String> _classes; // NOSONAR
	private final Map<String, String> _typeToGimlMapping; // NOSONAR
	private final Map<String, String> _typeToJavaMapping; // NOSONAR

	/**
	 * Instantiate a new template helper that load all template for future usage.
	 *
	 * @throws IllegalArgumentException
	 *             if class names cannot be dynamically read from {@link ClassName} interface.
	 * @throws IllegalAccessException
	 *             if class names cannot be dynamically read from {@link ClassName} interface.
	 */
	public TemplateHelper() throws IllegalArgumentException, IllegalAccessException {
		this._classDescription = new STGroupFile(CLASSDESCRIPTION_TEMPLATE_PATH);
		this._classDescriptionPackage = new STGroupFile(CLASSDESCRIPTIONPACKAGE_TEMPLATE_PATH);
		this._serviceDescriptionPackage = new STGroupFile(SERVICEDESCRIPTIONPACKAGE_TEMPLATE_PATH);
		this._classes = new HashMap<String, String>();
		this._typeToGimlMapping = new HashMap<String, String>();
		this._typeToJavaMapping = new HashMap<String, String>();

		StringRenderer renderer = new StringRenderer();
		this._classDescription.registerRenderer(String.class, renderer);
		this._classDescriptionPackage.registerRenderer(String.class, renderer);
		this._serviceDescriptionPackage.registerRenderer(String.class, renderer);

		this.initClasses();
		this.initTypeToGimlMapping();
		this.initTypeToJavaMapping();
	}

	/**
	 * Generates a {@link com.is2t.giml.description.Description} class content from the given description and Java
	 * source constants.
	 *
	 * @param description
	 *            content of the description class to create.
	 * @param packageName
	 *            the name of the package of the generated class.
	 * @param className
	 *            the generated class name.
	 * @return the content of generated class content.
	 */
	public String generateClassDescription(Description description, String packageName, String className) {
		Map<String, String> childMethodNames = new HashMap<String, String>();
		for (Child child : description.getChildren()) {
			String methodName = child.getPrefix();
			String childName = child.getName();
			if (child.hasEmptyName()) {
				childName = DEFAULT_ELEMENT;
			} else {
				methodName += childName.substring(0, 1).toUpperCase() + childName.substring(1);
			}
			childMethodNames.put(childName, methodName);
		}
		Map<String, Object> parameters = new HashMap<String, Object>();
		parameters.put(TEMPLATE_PARAMETER_DESCRIPTION, description);
		parameters.put(TEMPLATE_PARAMETER_CLASSES, this._classes);
		parameters.put(TEMPLATE_PARAMETER_TYPETOGIMLMAPPING, this._typeToGimlMapping);
		parameters.put(TEMPLATE_PARAMETER_TYPETOJAVAMAPPING, this._typeToJavaMapping);
		parameters.put(TEMPLATE_PARAMETER_PACKAGE, packageName);
		parameters.put(TEMPLATE_PARAMETER_CLASSNAME, className);
		parameters.put(TEMPLATE_PARAMETER_CHILDMETHODNAMES, childMethodNames);
		return this.generate(this._classDescription, TEMPLATE_CLASSDESCRIPTION, parameters);
	}

	/**
	 * Generates a {@link DescriptionPackage} class content from the given description class names and Java source
	 * constants.
	 *
	 * @param descriptionClassNames
	 *            list of previously created description classes.
	 * @param packageName
	 *            the name of the package of the generated class.
	 * @param className
	 *            the generated class name.
	 * @return the content of generated class content.
	 *
	 * @see #generateClassDescription(Description, String, String)
	 */
	public String generateClassDescriptionPackage(List<String> descriptionClassNames, String packageName,
			String className) {
		Map<String, Object> parameters = new HashMap<String, Object>();
		parameters.put(TEMPLATE_PARAMETER_DESCRIPTIONS, descriptionClassNames);
		parameters.put(TEMPLATE_PARAMETER_CLASSES, this._classes);
		parameters.put(TEMPLATE_PARAMETER_PACKAGE, packageName);
		parameters.put(TEMPLATE_PARAMETER_CLASSNAME, className);
		return this.generate(this._classDescriptionPackage, TEMPLATE_CLASSDESCRIPTIONPACKAGE, parameters);
	}

	/**
	 * Generates a service declaration for generated description package class.
	 *
	 * @param packageName
	 *            the name of the package of the generated description package class.
	 * @param className
	 *            the name of the generated description package class.
	 * @return the content of generated service.
	 */
	public String generateServiceDescriptionPackage(String packageName, String className) {
		Map<String, Object> parameters = new HashMap<String, Object>();
		parameters.put(TEMPLATE_PARAMETER_PACKAGE, packageName);
		parameters.put(TEMPLATE_PARAMETER_CLASSNAME, className);
		return this.generate(this._serviceDescriptionPackage, TEMPLATE_SERVICEDESCRIPTIONPACKAGE, parameters);
	}

	private String generate(STGroup group, String templateName, Map<String, Object> parameters) {
		ST template = group.getInstanceOf(templateName);
		for (Map.Entry<String, Object> parameter : parameters.entrySet()) {
			template.add(parameter.getKey(), parameter.getValue());
		}
		return template.render().concat(DOUBLE_BREAK_LINE);
	}

	private void initClasses() throws IllegalArgumentException, IllegalAccessException {
		Field[] declaredFields = ClassName.class.getDeclaredFields();
		for (Field field : declaredFields) {
			this._classes.put(field.getName(), (String) field.get(null));
		}
	}

	private void initTypeToGimlMapping() {
		String[] simpleMapping = new String[] { "byte", "short", "integer", "long", "float", "double", "boolean",
		"string" };
		for (String type : simpleMapping) {
			this._typeToGimlMapping.put(type, ClassName.GIML_TYPE.concat(".").concat(type.toUpperCase()));
		}
		this._typeToGimlMapping.put("int", ClassName.GIML_TYPE.concat(".INTEGER"));
		this._typeToGimlMapping.put("reference", ClassName.GIML_TYPE.concat(".REFERENCE"));
		this._typeToGimlMapping.put("code", ClassName.GIML_TYPE.concat(".CODE"));
		this._typeToGimlMapping.put(String.class.getCanonicalName(), ClassName.GIML_TYPE.concat(".STRING"));
	}

	private void initTypeToJavaMapping() {
		String[] simpleMapping = new String[] { "byte", "short", "int", "long", "float", "double", "boolean" };
		for (String type : simpleMapping) {
			this._typeToJavaMapping.put(type, type.toLowerCase());
		}
		this._typeToJavaMapping.put("integer", "int");
		this._typeToJavaMapping.put("reference", ClassName.STRING);
		// this._typeToJavaMapping.put(String.class.getCanonicalName(), ClassName.STRING);
	}

}
