/*
 * Java
 *
 * Copyright 2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.classpath;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * The data collected to the loaded classpath.
 */
public class ClasspathModel {

	private final Map<String, List<InstanceMethod>> instanceMethods;

	public ClasspathModel() {
		instanceMethods = new HashMap<>();
	}

	/**
	 *
	 * @param type
	 *            fullyQualifiedName of the type
	 * @param name
	 *            the simple method name
	 * @param desc
	 *            the classfile descriptor
	 */
	public void addInstanceMethod(String type, String name, String desc) {
		List<InstanceMethod> methodsVect = instanceMethods.get(name);
		if (methodsVect == null) {
			methodsVect = new ArrayList<ClasspathModel.InstanceMethod>(2);
			instanceMethods.put(name, methodsVect);
		}
		InstanceMethod m = new InstanceMethod(type, name, desc);
		methodsVect.add(m);
	}

	/**
	 * Get the list of instance methods with the given name and the specified number
	 * of parameters
	 *
	 * @param name
	 * @param nbParameters
	 * @return
	 */
	public InstanceMethod[] getAllInstanceMethods(String name, int nbParameters) {
		List<InstanceMethod> instanceMethodsSameName = instanceMethods.get(name);
		List<InstanceMethod> instanceMethodsSameNameAndNbParams = new ArrayList<>();
		for (InstanceMethod m : instanceMethodsSameName) {
			if (m.getParameterDescriptors().length == nbParameters) {
				instanceMethodsSameNameAndNbParams.add(m);
			}
		}
		return instanceMethodsSameNameAndNbParams
				.toArray(new InstanceMethod[instanceMethodsSameNameAndNbParams.size()]);
	}

	/**
	 * All classfile descriptors.
	 */
	public static class InstanceMethod {

		private final String type;
		private final String name;
		private final String desc;
		private String[] parameterDescriptors;
		private String returnDescriptors;

		public InstanceMethod(String type, String name, String desc) {
			this.type = type;
			this.name = name;
			this.desc = desc;

			initDesc(desc);
		}

		private void initDesc(String desc) {
			int ptr = 1;
			assert (desc.charAt(0) == '(');
			int endParams = desc.indexOf(')');
			assert endParams != -1;
			ArrayList<String> parameterDescriptorsVect = new ArrayList<>();
			while (ptr < endParams) {
				int endParam;
				switch (desc.charAt(ptr)) {
				case 'L': {
					endParam = desc.indexOf(';', ptr);
					assert endParam != -1;
					break;
				}
				default:
					endParam = ptr;
				}
				int nextParamPtr = endParam + 1;
				try {

					parameterDescriptorsVect.add(desc.substring(ptr, nextParamPtr));
				} catch (Throwable e) {
					e.printStackTrace();
				}
				ptr = nextParamPtr;
			}

			parameterDescriptors = parameterDescriptorsVect.toArray(new String[parameterDescriptorsVect.size()]);
			returnDescriptors = desc.substring(endParams + 1, desc.length());
		}

		/**
		 * Gets the desc.
		 *
		 * @return the desc.
		 */
		public String getDesc() {
			return desc;
		}

		/**
		 * Gets the name.
		 *
		 * @return the name.
		 */
		public String getName() {
			return name;
		}

		/**
		 * Gets the type.
		 *
		 * @return the type.
		 */
		public String getType() {
			return type;
		}

		/**
		 * Gets the parameterDescriptors.
		 *
		 * @return the parameterDescriptors.
		 */
		public String[] getParameterDescriptors() {
			return parameterDescriptors;
		}

		/**
		 * Gets the returnDescriptors.
		 *
		 * @return the returnDescriptors.
		 */
		public String getReturnDescriptors() {
			return returnDescriptors;
		}
	}
}
