/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.generator;

import java.io.StringWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.stringtemplate.v4.ST;

import com.is2t.giml.description.Description;
import com.is2t.giml.description.DescriptionFactory;
import com.is2t.giml.description.Type;
import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Node;

/**
 * This class {@code GIMLFactory} represents a GIML factory that holds a GIML
 * node informations to generate the factory source code.
 *
 */
public class GIMLFactory implements GIMLGeneratorConstants {

	private static final String ATTRIBUTE_CODE_SUFFIX = "code";

	/**
	 * The method name of the factory used as an unique identifier name of the
	 * factory.
	 */
	public final String uniqueID;

	/**
	 * The Description of the factory
	 */
	public final Description description;

	/**
	 * The type of the factory
	 */
	public final FactoryType factotyType;

	private int variableCount = 0;
	private final Node node;
	private boolean hasIdAttribute;
	private ST attributesList;
	private boolean needGIMLReader;
	/**
	 * GIML factory type
	 *
	 */
	public enum FactoryType {
		/**
		 * type for a factory that manages no child elements.
		 */
		NoElement,

		/**
		 * Factory that manages ordered child elements.
		 */
		OrderedElements,

		/**
		 * Factory that manages arrays of ordered child elements.
		 */
		OrderedArrayElements,

		/**
		 * type for a factory that manages unordered child elements.
		 */
		NotOrderedElements,

		/**
		 * type for a factory that manages arrays of unordered child elements.
		 */
		NotOrderedArrayElements,
	}

	/**
	 * Creates a new GIML factory for the given node.
	 *
	 * @param node
	 *            the node for which the factory will be created.
	 * @throws GIMLGeneratorException
	 *             if an error occurs.
	 */
	public GIMLFactory(Node node) throws GIMLGeneratorException {
		if (node == null) {
			throw new IllegalArgumentException();
		}
		this.node = node;
		this.description = DescriptionFactory.getDescription(node.getName());
		if (this.description == null) {
			throw new GIMLGeneratorException(String.format("no description found for '%s' node", node.getName()));
		}

		this.factotyType = getFactoryType();
		this.uniqueID = computeUniqueIdentifier();
	}

	/*
	 * Computes the unique identifier name for this factory
	 */
	private String computeUniqueIdentifier() {
		this.attributesList = GIMLGenerator.GimlSTGroup.getInstanceOf("AttributesList");
		StringBuilder uniqueId = new StringBuilder("GIMLFactory");
		uniqueId.append("_").append(this.node.getName().toLowerCase());
		Attribute[] attributes = GIMLGeneratorUtils.sortNodeAttributesByName(node);
		for (Attribute attribute : attributes) {
			String attributeName = attribute.getName();
			if (attributeName.equalsIgnoreCase(ID_ATTRIBUTE)) {
				this.hasIdAttribute = true;
			}
			uniqueId.append("_").append(attributeName.toLowerCase());
			this.attributesList.add("names", attributeName);
		}

		switch (factotyType) {
		case NoElement:
			// leaf
			break;
		case OrderedElements:
			appendNodeNames(node.getChildren(), uniqueId);
			break;
		case NotOrderedElements:
			// sort elements and change the order
			List<Node> sortedElements = GIMLGeneratorUtils.sortNodeChildrenByName(node);
			appendNodeNames(sortedElements, uniqueId);
			break;
		case OrderedArrayElements:
		case NotOrderedArrayElements:
			// uses the description elements name to manage elements as array
			Collection<String> descriptionElementsNames = this.description.getElements();
			String[] sortedDescriptionElementsNames = GIMLGeneratorUtils.sortIngnoreCase(descriptionElementsNames);
			appendNodeNames(sortedDescriptionElementsNames, uniqueId);
			// need GIMLReader to read the array element count
			needGIMLReader = true;

			break;
		default:
			throw new RuntimeException();
		}

		return uniqueId.toString();
	}

	private StringBuilder appendNodeNames(List<Node> elements, StringBuilder names) {
		for (Node element : elements) {
			names.append("_").append(element.getName().toLowerCase());
		}
		return names;
	}

	private StringBuilder appendNodeNames(String[] nodeNames, StringBuilder names) {
		for (String name : nodeNames) {
			names.append("_").append(name.toLowerCase());
		}
		return names;
	}

	/*
	 * Gets the factory type according to its description
	 */
	private FactoryType getFactoryType() {

		Collection<String> descriptionElements = this.description.getElements();
		int descriptionElementsCount = descriptionElements.size();

		if (descriptionElementsCount == 0) {
			// leaf node
			return FactoryType.NoElement;
		}

		Iterator<String> iterator = descriptionElements.iterator();
		if (descriptionElementsCount == 1) {
			String descElement = iterator.next();
			if (this.description.getElementMaxCardinality(descElement) == 1) {
				// 1 max child element
				return FactoryType.OrderedElements;
			} else {
				// multiple element with the same tag
				return FactoryType.OrderedArrayElements;
			}
		} else if (this.description.hasOrderedElements()) {
			// ordered elements with different tags
			return FactoryType.OrderedElements;
		} else {
			boolean elementsWithOneMaxCard = true;
			// checks if all elements has their cardinality to 1
			while (iterator.hasNext()) {
				String descElement = iterator.next();
				if (this.description.getElementMaxCardinality(descElement) > 1) {
					elementsWithOneMaxCard = false;
					break;
				}
			}
			if (elementsWithOneMaxCard) {
				// not ordered element and each element tag appears at most once
				return FactoryType.NotOrderedElements;
			}
			// not ordered element and each element tag may appears several
			// times
			return FactoryType.NotOrderedArrayElements;
		}
	}

	/**
	 * Generates the factory source code.
	 *
	 * @return the StringTeample object representing the code.
	 * @throws GIMLGeneratorException
	 *             if a an error occurs during the code generation.
	 */
	public ST generate() throws GIMLGeneratorException {
		String nodeName = this.node.getName();
		ST factoryTemplate = GIMLGenerator.GimlSTGroup.getInstanceOf("GIMLFactory");

		// generate the receiver variableName
		String receiverVariableName = computeNextVariable(nodeName);
		// add the computed factory class name
		factoryTemplate.add("methodName", this.uniqueID);

		factoryTemplate.add("node", nodeName.toLowerCase());
		factoryTemplate.add("receiverVariable", receiverVariableName);
		factoryTemplate.add("hasIdAttribute", this.hasIdAttribute);
		factoryTemplate.add("attributesList", this.attributesList);

		// generate initialize code
		generateInitializationCode(factoryTemplate, receiverVariableName);

		// process elements code generation
		generateChildElementsCode(factoryTemplate, receiverVariableName);

		factoryTemplate.add("needGIMLReader", this.needGIMLReader);
		return factoryTemplate;

	}

	private void generateInitializationCode(ST factoryTemplate, String receiverVariableName)
			throws GIMLGeneratorException {
		StringWriter writer = new StringWriter();
		Map<String, String> attributeVariableMap = new HashMap<>();

		// generate code for attribute reading
		String[] attributes = GIMLGeneratorUtils.sortAttributesNamesByName(this.node);
		ST attributesTemplate = generateAttributes(this.node.getName(), attributes, attributeVariableMap, false,
				description.isRoot());
		ST constructorCodeAttributesTemplate = generateConstructorExecCodeAttributes(this.node.getName(), attributes,
				attributeVariableMap, false, description.isRoot());

		factoryTemplate.add("attributes", attributesTemplate);
		factoryTemplate.add("constructorCodeAttributes", constructorCodeAttributesTemplate);

		// generate initialization code
		this.description.generateInitialization(receiverVariableName, attributeVariableMap, writer);
		// add the initialization description code
		factoryTemplate.add("initializationDescriptionCode", writer.toString());

	}

	private ST generateAttributes(String nodeName, String[] attributes, Map<String, String> attributesMap,
			boolean isElementNode, boolean isRoot) throws GIMLGeneratorException {
		ST attributesTemplate = GIMLGenerator.GimlSTGroup.getInstanceOf("GenerateAttributes");
		// get the node attributes infos (name, type, reader method etc.) and
		// fill the template
		for (String attributeName : attributes) {
			if (attributeName.equalsIgnoreCase(ID_ATTRIBUTE)) {
				if (!isElementNode && !isRoot) {
					// ignore id attribute
					continue;
				}
			}

			Type type = GIMLGenerator.getAttributeType(nodeName, attributeName, this.description, isElementNode);
			String gimlReaderMethodName = getGIMLReaderMethodName(type);
			String attributeVariableName = computeNextVariable(attributeName);

			// add attribute infos
			attributesTemplate.addAggr("attributes.{node, name, type, variableName, GIMLReaderMethod}", nodeName,
					attributeName, type.getTypeName(), attributeVariableName, gimlReaderMethodName);

			attributesMap.put(attributeName.toLowerCase(), attributeVariableName);

		}
		if (this.needGIMLReader == false && attributes.length > 0) {
			// need GIMLReader if there is at least one attribute to read
			this.needGIMLReader = true;
		}
		return attributesTemplate;
	}

	private ST generateConstructorExecCodeAttributes(String nodeName, String[] attributes,
			Map<String, String> attributesMap, boolean isElementNode, boolean isRoot) throws GIMLGeneratorException {
		ST attributesTemplate = GIMLGenerator.GimlSTGroup.getInstanceOf("GenerateConstructorExecCodeAttributes");
		// get the node attributes infos (name, type, reader method etc.) and
		// fill the template
		for (String attributeName : attributes) {
			if (attributeName.equalsIgnoreCase(ID_ATTRIBUTE)) {
				if (!isElementNode && !isRoot) {
					// ignore id attribute
					continue;
				}
			}
			Type type = GIMLGenerator.getAttributeType(nodeName, attributeName, this.description, isElementNode);
			if (type.equals(Type.CODE) && attributeName.toLowerCase().endsWith(ATTRIBUTE_CODE_SUFFIX)) {
				String derivedAttributeName = attributeName.substring(0,
						attributeName.length() - ATTRIBUTE_CODE_SUFFIX.length());
				Type derivedType = GIMLGenerator.getAttributeType(nodeName, derivedAttributeName, this.description,
						isElementNode);
				if (derivedType == null) {
					continue; // this is not a derived attribute (most likely a hook ending with "Code")
				}

				String attributeVariableName = attributesMap.get(attributeName.toLowerCase());
				String attributeVariableNameResult = computeNextVariable(attributeName);
				// register the derived attribute variable mapping
				attributesMap.put(derivedAttributeName.toLowerCase(), attributeVariableNameResult);

				// add attribute infos
				attributesTemplate.addAggr(
						"constructorCodeAttributes.{node, name, type, variableName, variableNameResult, convertMethodName}",
						nodeName, attributeName, derivedType.getTypeName(), attributeVariableName,
						attributeVariableNameResult, getJSConvertMethodName(derivedType));
			}

		}
		return attributesTemplate;
	}

	private Object getJSConvertMethodName(Type derivedType) {
		String getTypeSuffix = derivedType.getTypeName();
		int lastDot = getTypeSuffix.lastIndexOf('.');
		if (lastDot != -1) {
			getTypeSuffix = getTypeSuffix.substring(lastDot + 1, getTypeSuffix.length());
		}
		return "com.is2t.giml.js.GIMLJavaScriptCode.JS_To" + getTypeSuffix;
	}

	private void generateArrayElements(ST factoryTemplate, String receiverVariable) throws GIMLGeneratorException {
		factoryTemplate.add("hasArrayElements", true);
		String[] elementNames = GIMLGeneratorUtils.sortIngnoreCase(this.description.getElements());
		for (String elementName : elementNames) {
			Collection<String> attributeNames = this.description.getElementAttributes(elementName);
			String[] attributeNamesArray = GIMLGeneratorUtils.sortIngnoreCase(attributeNames);
			generateElementIntern(elementName, attributeNamesArray, factoryTemplate, receiverVariable);
		}
	}

	private void generateElements(List<Node> elements, ST factoryTemplate, String receiverVariableName)
			throws GIMLGeneratorException {

		for (Node element : elements) {
			String[] attributes = GIMLGeneratorUtils.sortAttributesNamesByName(element);
			String elementName = element.getName();
			generateElementIntern(elementName, attributes, factoryTemplate, receiverVariableName);

		}
	}

	private void generateElementIntern(String elementName, String[] attributes, ST factoryTemplate,
			String receiverVariableName) throws GIMLGeneratorException {
		Map<String, String> elementAttributesMap = new HashMap<>();
		// generate element attributes
		ST attributesTemplate = generateAttributes(elementName, attributes, elementAttributesMap, true, false);


		String elementType = this.description.getElementType(elementName).getTypeName();
		String elementVariableName = computeNextVariable(elementName);
		StringWriter writer = new StringWriter();

		// generate the current element description code
		this.description.generateChild(receiverVariableName, elementName, elementAttributesMap, elementVariableName,
				writer);

		// fill the factory template with element infos
		factoryTemplate.addAggr(
				"elements.{name, type, variableName, attributes, descriptionCode}", elementName, elementType,
				elementVariableName, attributesTemplate,
				writer.toString());
	}

	private void generateChildElementsCode(ST factoryTemplate, String receiverVariableName)
			throws GIMLGeneratorException {

		switch (this.factotyType) {
		case NoElement:
			// leaf
			return;
		case OrderedElements:
			generateElements(node.getChildren(), factoryTemplate, receiverVariableName);
			break;
		case OrderedArrayElements:
			// ex : List
			generateArrayElements(factoryTemplate, receiverVariableName);
			break;
		case NotOrderedElements:
			// ex : Dock
			List<Node> sortedElements = GIMLGeneratorUtils.sortNodeChildrenByName(node);
			generateElements(sortedElements, factoryTemplate, receiverVariableName);
			break;
		case NotOrderedArrayElements:
			generateArrayElements(factoryTemplate, receiverVariableName);
			break;
		default:
			throw new RuntimeException();
		}

	}

	private String computeNextVariable(String variableSuffix) {
		return new StringBuffer(FACTORY_VARIABLE_PREFIX).append(++this.variableCount).append("_").append(variableSuffix)
				.toString();
	}

	@Override
	public boolean equals(Object o) {

		if (o == null || !(o instanceof GIMLFactory)) {
			return false;
		}

		GIMLFactory that = (GIMLFactory) o;
		return this.uniqueID.equals(that.uniqueID);
	}

	private final String getGIMLReaderMethodName(Type type) {
		switch (type.getBasetype()) {
		case Boolean:
			return "readBoolean";
		case Byte:
		case Short:
		case Integer:
		case Code:
			return "readVarSInt";
		case Long:
			return "readVarLong";
		case Float:
			return "readFloat";
		case Double:
			return "readDouble";
		case String:
		case RefId:
			return "readString";
		default:
			throw new RuntimeException(" No GIML reader method found for the type " + type.getBasetype().getName());
		}
	}


}
