/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.generator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Element;

import com.is2t.giml.xml.generator.exception.XMLGeneratorException;
import com.is2t.giml.xml.generator.utils.FileUtils;
import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.Template;

public class XMLGenerator {

	private final Document _document;
	private XMLGeneratorOptions _options;

	/**
	 * Creates a generator with default options.
	 *
	 * @param document
	 *            the document to generate.
	 * @throws NullPointerException
	 *             if the given document is {@code null}.
	 */
	public XMLGenerator(Document document) {
		if (document == null) {
			throw new NullPointerException();
		}
		this._document = document;
		this._options = new XMLGeneratorOptions();
	}

	/**
	 * Gets parser options.
	 *
	 * @return parser options.
	 */
	public XMLGeneratorOptions getOptions() {
		return this._options;
	}

	/**
	 * Sets options for this parser.
	 *
	 * @param options
	 *            new options.
	 * @throws IllegalArgumentException
	 *             if given options are <code>null</code>.
	 */
	public void setOptions(XMLGeneratorOptions options) throws IllegalArgumentException {
		if (options == null) {
			throw new IllegalArgumentException("Parser options cannot be null");
		}
		this._options = options;
	}

	/**
	 * Generates the document to a file along with its dependencies recursively.
	 *
	 * @param directory
	 *            the directory to generate the documents to.
	 * @throws XMLGeneratorException
	 *             if an error occurs while generating the XML.
	 * @throws IllegalArgumentException
	 *             if the given directory does not exist or does not denote a directory.
	 */
	public void generateRecursively(File directory) throws XMLGeneratorException {
		FileUtils.checkValidDirectory(directory);
		try {
			generateRecursively(this._document, directory, new ArrayList<Document>());
		} catch (FileNotFoundException | TransformerFactoryConfigurationError | ParserConfigurationException
				| TransformerException e) {
			throw new XMLGeneratorException(e.getMessage(), e);
		}
	}

	private void generateRecursively(Document document, File directory, List<Document> generatedDocuments)
			throws FileNotFoundException, TransformerFactoryConfigurationError, ParserConfigurationException,
			TransformerException {
		generatedDocuments.add(document);
		generate(document,
				new FileOutputStream(new File(directory, document.getId() + this._options.getXMLFileExtension())));
		List<Document> referencedDocuments = document.getReferencedDocuments();
		for (Document referencedDocument : referencedDocuments) {
			if (!generatedDocuments.contains(referencedDocument)) {
				generateRecursively(referencedDocument, directory, generatedDocuments);
			}
		}
	}

	/**
	 * Generates the document to a file.
	 *
	 * @param file
	 *            the file to generate the document to.
	 * @throws XMLGeneratorException
	 *             if an error occurs while generating the XML.
	 * @throws IllegalArgumentException
	 *             if the given file does not denote a file.
	 */
	public void generate(File file) throws XMLGeneratorException {
		if (file.exists() && file.isDirectory()) {
			throw new IllegalArgumentException(String.format("'%s' is a directory.", file.getAbsolutePath()));
		}
		try {
			generate(this._document, new FileOutputStream(file));
		} catch (TransformerFactoryConfigurationError | ParserConfigurationException | TransformerException
				| FileNotFoundException e) {
			throw new XMLGeneratorException(e.getMessage(), e);
		}
	}

	/**
	 * Generates the document to an output stream.
	 *
	 * @param outputStream
	 *            the file to generate the document to.
	 * @throws XMLGeneratorException
	 *             if an error occurs while generating the XML.
	 * @throws NullPointerException
	 *             if the given output stream is <code>null</code>.
	 */
	public void generate(OutputStream outputStream) throws XMLGeneratorException {
		try {
			generate(this._document, outputStream);
		} catch (TransformerFactoryConfigurationError | ParserConfigurationException | TransformerException
				| FileNotFoundException e) {
			throw new XMLGeneratorException(e.getMessage(), e);
		}
	}

	private void generate(Document document, OutputStream outputStream) throws TransformerFactoryConfigurationError,
			ParserConfigurationException, TransformerException, FileNotFoundException {
		org.w3c.dom.Document domDocument = transformDocument(document);
		domDocument.setXmlStandalone(true);

		Transformer transformer = TransformerFactory.newInstance().newTransformer();
		DOMSource source = new DOMSource(domDocument);
		StreamResult result = new StreamResult(outputStream);
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");
		transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
		transformer.setOutputProperty(OutputKeys.STANDALONE, "yes");
		transformer.transform(source, result);
	}

	private org.w3c.dom.Document transformDocument(Document document) throws ParserConfigurationException {
		DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
		org.w3c.dom.Document domDocument = documentBuilder.newDocument();

		// Generate nodes recursively.
		Element rootElement = createElement(document, domDocument);

		XMLGeneratorOptions options = getOptions();
		org.w3c.dom.Node firstChild = rootElement.getFirstChild();

		// Generate imports.
		String importTagName = options.getImportTagName();
		String importTagAttributeName = options.getImportTagAttributeName();
		List<Document> referencedDocuments = document.getReferencedDocuments();
		for (Document referencedDocument : referencedDocuments) {
			Element importElement = domDocument.createElement(importTagName);
			importElement.setAttribute(importTagAttributeName, referencedDocument.getId());
			rootElement.insertBefore(importElement, firstChild);
		}

		// Generate templates.
		String templateTagName = options.getTemplateTagName();
		String templateTagAttributeName = options.getTemplateTagAttributeName();
		List<Template> templates = document.getTemplates();
		for (Template template : templates) {
			Element templateElement = domDocument.createElement(templateTagName);
			templateElement.setAttribute(templateTagAttributeName, template.getName());
			generateChildren(template.getTemplateNode(), domDocument, templateElement);
			rootElement.insertBefore(templateElement, firstChild);
		}

		domDocument.appendChild(rootElement);

		return domDocument;
	}

	private Element createElement(Document document, org.w3c.dom.Document domDocument) {
		Element rootElement = createElement((Node) document, domDocument);
		rootElement.setAttribute(this._options.getDocumentIdAttributeName(), document.getId());
		return rootElement;
	}

	private Element createElement(Node node, org.w3c.dom.Document domDocument) {
		Element element = domDocument.createElement(node.getName());
		Set<Attribute> attributes = node.getAttributes();
		for (Attribute attribute : attributes) {
			element.setAttribute(attribute.getName(), attribute.getValue());
		}
		generateChildren(node, domDocument, element);
		return element;
	}

	private void generateChildren(Node node, org.w3c.dom.Document domDocument, Element element) {
		List<Node> children = node.getChildren();
		for (Node child : children) {
			Element childElement = createElement(child, domDocument);
			element.appendChild(childElement);
		}
	}

}
