/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.generator.test;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.junit.AfterClass;
import org.junit.Ignore;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Node;

/**
 * The class {@code AbstractTest} provides utility methods for all unit test classes.
 */
@Ignore
public abstract class AbstractTest {

	private static final String TMP_SUFFIX = "XMLGeneratorTest";
	private final static List<File> FILE_TO_DELETE = new ArrayList<File>();
	private final static List<File> FOLDER_TO_DELETE = new ArrayList<File>();

	/**
	 * Empty set of attributes.
	 */
	protected static final Set<Attribute> NO_ATTRIBUTE = new HashSet<Attribute>();

	/**
	 * Creates an attribute set from given inputs.
	 *
	 * @param attributeDef
	 *            a name-value array of attribute to create.
	 * @return the set of attributes.
	 */
	protected static Set<Attribute> createAttributeSet(String... attributeDef) {
		Set<Attribute> attributes = new HashSet<Attribute>();
		for (int a = 0; a < attributeDef.length; a += 2) {
			attributes.add(new Attribute(attributeDef[a], attributeDef[a + 1]));
		}
		return attributes;
	}

	/**
	 * Creates a node from given inputs.
	 *
	 * @param name
	 *            the node tag name.
	 * @param attributes
	 *            the node attributes.
	 * @param children
	 *            the node children.
	 * @return the created node.
	 */
	protected static Node createNode(String name, Set<Attribute> attributes, Node... children) {
		List<Node> childNodes = new ArrayList<Node>();
		for (Node child : children) {
			childNodes.add(child);
		}
		return new Node(name, attributes, childNodes);
	}

	/**
	 * Automatically delete all created temporary files and folders in this test class.
	 */
	@AfterClass
	public static void deleteAllFiles() {
		for (File file : FILE_TO_DELETE) {
			file.delete();
		}
		FILE_TO_DELETE.clear();
		for (File folder : FOLDER_TO_DELETE) {
			deleteDirectoriesAndFiles(folder);
		}
		FOLDER_TO_DELETE.clear();
	}

	/**
	 * Creates a temporary directory.
	 *
	 * @param prefix
	 *            the prefix of the created temporary directory.
	 * @return the created temporary directory.
	 * @throws IOException
	 *             if an IO error occurred.
	 */
	protected static File createTempDir(String prefix) throws IOException {
		File f = File.createTempFile(prefix, TMP_SUFFIX);
		if (!f.delete()) {
			throw new IOException(f.getAbsolutePath());
		}
		if (!f.mkdirs()) {
			throw new IOException(f.getAbsolutePath());
		}
		// FOLDER_TO_DELETE.add(f);
		return f;
	}

	/**
	 * Creates a temporary file.
	 *
	 * @param prefix
	 *            the prefix of the created temporary file.
	 * @return the created temporary file.
	 * @throws IOException
	 *             if an IO error occurred.
	 */
	protected static File createTempFile(String prefix) throws IOException {
		File tempFile = File.createTempFile(prefix, TMP_SUFFIX);
		FILE_TO_DELETE.add(tempFile);
		return tempFile;
	}

	private static void deleteDirectoriesAndFiles(File basedir) {
		try {
			if (basedir.exists()) {
				File[] files = basedir.listFiles();
				if (files == null) {
					return;
				}
				for (File file : files) {
					if (file.isDirectory()) {
						deleteDirectoriesAndFiles(file);
					}
					if (!file.delete()) {
						throw new IOException(String.format("Cannot delete file '%s'.", file.getAbsolutePath()));
					}
				}
				if (!basedir.delete()) {
					throw new IOException(String.format("Cannot delete directory '%s'.", basedir.getAbsolutePath()));
				}
			}
		} catch (IOException e) {
			// we don't care, just for test!
		}
	}

}
