/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model;

import java.util.Objects;

/**
 * The <code>Attribute</code> class represents a {@link Node} attribute. An attribute is a couple <code>name</code>-
 * <code>value</code>.
 *
 * @see Node
 */
public class Attribute {

	private final String _name; // NOSONAR
	private final String _value; // NOSONAR

	/**
	 * Create a new attribute with given values.
	 *
	 * @param name
	 *            the name of the attribute
	 * @param value
	 *            the value of this attribute
	 *
	 * @throws IllegalArgumentException
	 *             if given name or value are <code>null</code> or empty.
	 */
	public Attribute(String name, String value) throws IllegalArgumentException {
		if (name == null || name.isEmpty()) {
			throw new IllegalArgumentException("Attribute name cannot be null or empty");
		}
		if (value == null || value.isEmpty()) {
			throw new IllegalArgumentException("Attribute value cannot be null or empty");
		}

		this._name = name;
		this._value = value;
	}

	/**
	 * Gets the name of this attribute.
	 *
	 * @return the name of this attribute.
	 */
	public String getName() {
		return this._name;
	}

	/**
	 * Gets the value of this attribute as a {@link String}.
	 *
	 * @return the string value of this attribute.
	 */
	public String getValue() {
		return this._value;
	}

	/**
	 * Gets the value of this attribute, as an integer value.
	 *
	 * @return the {@link Integer} value of this attribute.
	 *
	 * @throws NumberFormatException
	 *             if the value of this attribute does not represent an integer.
	 */
	public Integer getIntegerValue() throws NumberFormatException {
		return Integer.valueOf(Integer.parseInt(this._value));
	}

	/**
	 * Gets the value of this attribute, as a float value.
	 *
	 * @return the {@link Float} value of this attribute.
	 *
	 * @throws NumberFormatException
	 *             if the value of this attribute does not represent a float.
	 */
	public Float getFloatValue() throws NumberFormatException {
		return Float.valueOf(Float.parseFloat(this._value));
	}

	/**
	 * Gets the value of this attribute, as a double value.
	 *
	 * @return the {@link Double} value of this attribute.
	 *
	 * @throws NumberFormatException
	 *             if the value of this attribute does not represent a double.
	 */
	public Double getDoubleValue() throws NumberFormatException {
		return Double.valueOf(Double.parseDouble(this._value));
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof Attribute)) {
			return false;
		}
		Attribute other = (Attribute) obj;
		return this._name.equalsIgnoreCase(other._name) && this._value.equals(other._value);

	}

	@Override
	public int hashCode() {
		return Objects.hash(this._name, this._value);
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();

		builder.append("A[");
		builder.append(this._name);
		builder.append('=');
		builder.append(this._value);
		builder.append(']');

		return builder.toString();
	}

}
