/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * The <code>Node</code> class represents a node in a tree-like description. A node
 * <ul>
 * <li>has a name,</li>
 * <li>can have attributes,</li>
 * <li>can have children, i.e. child nodes</li>
 * </ul>
 */
public class Node {

	/**
	 * Name of this node.
	 */
	protected String _name; // NOSONAR
	/**
	 * List of declared attributes of this node.
	 */
	protected Set<Attribute> _attributes; // NOSONAR

	/**
	 * List of children of this node.
	 */
	protected List<Node> _children; // NOSONAR

	/**
	 * Create a new node with given name and no attribute.
	 *
	 * @param name
	 *            the name of this node.
	 * @throws IllegalArgumentException
	 *             if the given name is empty or <code>null</code>.
	 */
	public Node(String name) throws IllegalArgumentException {
		if (name == null) {
			throw new IllegalArgumentException("Node name cannot be null");
		}
		if (name.isEmpty()) {
			throw new IllegalArgumentException("Node name cannot be empty");
		}
		this._name = name;
		this._attributes = new HashSet<Attribute>();
		this._children = new ArrayList<Node>();
	}

	/**
	 * Create a new node with given name and attributes.
	 *
	 * @param name
	 *            the name of this node.
	 * @param attributes
	 *            declared attributes of the node to create.
	 * @throws IllegalArgumentException
	 *             if the given name is empty, or if given name or attribute set are <code>null</code>.
	 */
	public Node(String name, Set<Attribute> attributes) throws IllegalArgumentException {
		this(name);
		if (attributes == null) {
			throw new IllegalArgumentException("Node attributes cannot be null");
		}
		this._attributes.addAll(attributes);
	}

	/**
	 * Create a new node with given name and attributes.
	 *
	 * @param name
	 *            the name of this node.
	 * @param attributes
	 *            declared attributes of the node to create.
	 * @param children
	 *            children of this node.
	 * @throws IllegalArgumentException
	 *             if the given name is empty, or if given name, attribute set or child list are <code>null</code>.
	 */
	public Node(String name, Set<Attribute> attributes, List<Node> children) throws IllegalArgumentException {
		this(name, attributes);
		if (children == null) {
			throw new IllegalArgumentException("Node children cannot be null");
		}
		this._children.addAll(children);
	}

	/**
	 * Gets if this node contains a child with the given name.
	 *
	 * @param childName
	 *            the child name to find.
	 * @return <code>true</code> if this node contains a child with the given name, <code>false</code> otherwise. Also
	 *         returns <code>false</code> if the given child name is <code>null</code>.
	 */
	public boolean hasChild(String childName) {
		return !this.getChildrenByName(childName).isEmpty();
	}

	/**
	 * Gets the tag name of this node.
	 *
	 * @return the tag name of this node.
	 */
	public final String getName() {
		return this._name;
	}

	/**
	 * Gets the {@link Attribute} corresponding to the given attribute name.
	 *
	 * @param attributeName
	 *            the name of the attribute to get.
	 * @return the attribute associated to the given name, or <code>null</code> if the given attribute does not exist.
	 */
	public final Attribute getAttribute(String attributeName) {
		for (Attribute attribute : this._attributes) {
			if (attribute.getName().equalsIgnoreCase(attributeName)) {
				return attribute;
			}
		}
		return null;
	}

	/**
	 * Gets the value of the given attribute.
	 *
	 * @param attributeName
	 *            the name of the attribute to get the value of.
	 * @return the value of the given attribute as a {@link String}, or <code>null</code> if the given attribute does
	 *         not exist.
	 */
	public final String getAttributeValue(String attributeName) {
		Attribute attribute = this.getAttribute(attributeName);
		if (attribute != null) {
			return attribute.getValue();
		}

		return null;
	}

	/**
	 * Gets the value of the given attribute, as an integer value.
	 *
	 * @param attributeName
	 *            the name of the attribute to get the value of.
	 * @return the value of the given attribute as a {@link Integer}, or <code>null</code> if the given attribute does
	 *         not exist.
	 *
	 * @throws NumberFormatException
	 *             if the value of the given attribute does not represent an integer.
	 */
	public final Integer getIntegerAttributeValue(String attributeName) throws NumberFormatException {
		Attribute attribute = this.getAttribute(attributeName);
		if (attribute != null) {
			return attribute.getIntegerValue();
		}

		return null;
	}

	/**
	 * Gets the value of the given attribute, as a float value.
	 *
	 * @param attributeName
	 *            the name of the attribute to get the value of.
	 * @return the value of the given attribute as a {@link Float}, or <code>null</code> if the given attribute does not
	 *         exist.
	 *
	 * @throws NumberFormatException
	 *             if the value of the given attribute does not represent a float.
	 */
	public final Float getFloatAttributeValue(String attributeName) throws NumberFormatException {
		Attribute attribute = this.getAttribute(attributeName);
		if (attribute != null) {
			return attribute.getFloatValue();
		}

		return null;
	}

	/**
	 * Gets the value of the given attribute, as a double value.
	 *
	 * @param attributeName
	 *            the name of the attribute to get the value of.
	 * @return the value of the given attribute as a {@link Double}, or <code>null</code> if the given attribute does
	 *         not exist.
	 *
	 * @throws NumberFormatException
	 *             if the value of the given attribute does not represent a double.
	 */
	public final Double getDoubleAttributeValue(String attributeName) throws NumberFormatException {
		Attribute attribute = this.getAttribute(attributeName);
		if (attribute != null) {
			return attribute.getDoubleValue();
		}

		return null;
	}

	/**
	 * Gets the number of declared attributes of this node.
	 *
	 * @return the number of declared attributes of this node.
	 */
	public final int getAttributeCount() {
		return this._attributes.size();
	}

	/**
	 * <p>
	 * Gets a read-only set of declared attributes of this node.
	 * </p>
	 * <p>
	 * Any attempt to modify this set will result in a {@link UnsupportedOperationException}.
	 * </p>
	 *
	 * @return a possibly empty set of this node declared attributes.
	 */
	public final Set<Attribute> getAttributes() {
		return Collections.unmodifiableSet(this._attributes);
	}

	/**
	 * <p>
	 * Gets a read-only set of declared attribute names of this node.
	 * </p>
	 * <p>
	 * Any attempt to modify this set will result in a {@link UnsupportedOperationException}.
	 * </p>
	 *
	 * @return a possibly empty set of this node declared attribute names.
	 */
	public final Set<String> getAttributeNames() {
		Set<String> attributeNames = new HashSet<String>();
		for (Attribute attribute : this._attributes) {
			attributeNames.add(attribute.getName());
		}
		return Collections.unmodifiableSet(attributeNames);
	}

	/**
	 * Gets the number of children of this node.
	 *
	 * @return the number of children of this node.
	 */
	public int getChildCount() {
		return this._children.size();
	}

	/**
	 * Gets the list of node children that have the given name.
	 *
	 * @param name
	 *            the name of children to find.
	 * @return a possibly empty list of this node children that match the given name. If given name is <code>null</code>
	 *         , an empty list is returned.
	 */
	public List<Node> getChildrenByName(String name) {
		List<Node> matchingNodes = new ArrayList<Node>();
		if (name != null) {
			for (Node child : this._children) {
				if (child.getName().equalsIgnoreCase(name)) {
					matchingNodes.add(child);
				}
			}
		}
		return matchingNodes;
	}

	/**
	 * <p>
	 * Gets a read-only list of this node children.
	 * </p>
	 * <p>
	 * Any attempt to modify this list will result in a {@link UnsupportedOperationException}.
	 * </p>
	 *
	 * @return a possibly empty list of this node children.
	 */
	public List<Node> getChildren() {
		return Collections.unmodifiableList(this._children);
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof Node)) {
			return false;
		}

		Node other = (Node) obj;
		return this._name.equalsIgnoreCase(other._name) && this._attributes.equals(other._attributes)
				&& this._children.equals(other._children);
	}

	@Override
	public int hashCode() {
		return Objects.hash(this._name, this._attributes, this._children);
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();

		builder.append("N[");
		builder.append("name=");
		builder.append(this._name);
		if (this.getAttributeCount() > 0) {
			builder.append(',');
			builder.append("attributes=");
			builder.append('{');
			Iterator<Attribute> iterator = this._attributes.iterator();
			while (iterator.hasNext()) {
				builder.append(iterator.next().toString());
				if (iterator.hasNext()) {
					builder.append(',');
				}
			}
			builder.append('}');
		}
		if (this.getChildCount() > 0) {
			builder.append(',');
			builder.append("children=");
			builder.append('{');
			Iterator<Node> iterator = this._children.iterator();
			while (iterator.hasNext()) {
				builder.append(iterator.next().getName());
				if (iterator.hasNext()) {
					builder.append(',');
				}
			}
			builder.append('}');
		}
		builder.append(']');

		return builder.toString();
	}
}
