/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model.test.template;

import java.util.List;
import java.util.Set;

import org.junit.Test;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.Template;
import com.is2t.giml.xml.model.test.AbstractTest;

import junit.framework.Assert;

/**
 * Unit tests for {@link XMLParserTemplate} class.
 */
public class TemplateTest extends AbstractTest {

	private static final String TEMPLATE_NAME = "my-template";

	private Node createTemplateNodeWithDefaultOptionValues(String name, Node... children) {
		Set<Attribute> attributeSet = createAttributeSet(TEMPLATE_ATTR_NAME, name);
		return createNode(TEMPLATE_TAG_NAME, attributeSet, children);
	}

	// Invalid template tag name

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateTagName() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet(TEMPLATE_ATTR_NAME, TEMPLATE_NAME);
		Node invalidNode = createNode("invalid-template-tag", attributeSet);
		createTemplate(invalidNode);
	}

	// Invalid template attribute name

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateAttributeName() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("invalid-attr-name", TEMPLATE_NAME);
		Node invalidNode = createNode(TEMPLATE_TAG_NAME, attributeSet);
		createTemplate(invalidNode);
	}

	// Invalid child count

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateTooMuchChildren() throws NullPointerException, IllegalArgumentException {
		Node node1 = createNode(TEMPLATE_TAG_NAME, NO_ATTRIBUTE);
		Node node2 = createNode(TEMPLATE_TAG_NAME, NO_ATTRIBUTE);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, node1, node2);
		createTemplate(templateNode);
	}

	// Invalid Parameters in first child

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter1() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam("param1 1"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter2() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam("param1" + TEMPLATE_PREFIX));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter3() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam("param1" + TEMPLATE_SUFFIX));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter4() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam(TEMPLATE_SUFFIX + "param1"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter5() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam(TEMPLATE_PREFIX + "param1"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter6() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text",
				createTemplateParam("param1" + createTemplateParam("param2")));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter7() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text",
				createTemplateParam("param1" + createTemplateParam("param2") + "param3"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	// Invalid parameters in nested children

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateParameter8() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> subChildAttributeSet = createAttributeSet("text", createTemplateParam("param1 1"));
		Node subChild = createNode("node2", subChildAttributeSet);
		Node child = createNode("node1", NO_ATTRIBUTE, subChild);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		createTemplate(templateNode);
	}

	// Invalid template instantiation

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateInstantiationWrongName() throws NullPointerException, IllegalArgumentException {
		Node child = createNode("node1", NO_ATTRIBUTE);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Node instantiation = createNode(TEMPLATE_NAME + "aaa", NO_ATTRIBUTE);
		template.instantiate(instantiation);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateInstantiationMissingParameter()
			throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam("param1"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Node instantiation = createNode(TEMPLATE_NAME, NO_ATTRIBUTE);
		template.instantiate(instantiation);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateInstantiationTooMuchParameters()
			throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam("param1"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Set<Attribute> attributeSetInvocation = createAttributeSet("param1", "value1", "param2", "value2");
		Node instantiation = createNode(TEMPLATE_NAME, attributeSetInvocation);
		template.instantiate(instantiation);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateInstantiationMissingContent() throws NullPointerException, IllegalArgumentException {
		Node child = createNode("node1", NO_ATTRIBUTE,
				createNode(TEMPLATE_CONTENT_TAG, createAttributeSet(TEMPLATE_CONTENT_ATTR_NAME, "myContent")));
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Node instantiation = createNode(TEMPLATE_NAME, NO_ATTRIBUTE);
		template.instantiate(instantiation);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testInvalidTemplateInstantiationTooMuchContent() throws NullPointerException, IllegalArgumentException {
		Node child = createNode("node1", NO_ATTRIBUTE,
				createNode(TEMPLATE_CONTENT_TAG, createAttributeSet(TEMPLATE_CONTENT_ATTR_NAME, "myContent")));
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Node instantiation = createNode(TEMPLATE_NAME, NO_ATTRIBUTE,
				createNode(TEMPLATE_CONTENT_TAG,
						createAttributeSet(TEMPLATE_CONTENT_INSTANTIATION_ATTR_NAME, "myContent"),
						createNode("node2", NO_ATTRIBUTE)),
				createNode(TEMPLATE_CONTENT_TAG,
						createAttributeSet(TEMPLATE_CONTENT_INSTANTIATION_ATTR_NAME, "myContent2"),
						createNode("node3", NO_ATTRIBUTE)));
		template.instantiate(instantiation);
	}

	// Valid template

	@Test
	public void testTotallyEmptyTemplate() throws NullPointerException, IllegalArgumentException {
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME);
		Template template = createTemplate(templateNode);

		Assert.assertEquals(template.getName(), TEMPLATE_NAME);
		Assert.assertEquals(template.getParameterCount(), 0);
		Assert.assertEquals(template.getParameters().size(), 0);
		Assert.assertEquals(template.getTemplateTagName(), TEMPLATE_TAG_NAME);
		Assert.assertEquals(template.getTemplateTagAttributeName(), TEMPLATE_ATTR_NAME);
		Assert.assertEquals(template.getParameterPrefix(), TEMPLATE_PREFIX);
		Assert.assertEquals(template.getParameterSuffix(), TEMPLATE_SUFFIX);
		Assert.assertNull(template.getRoot());
	}

	@Test
	public void testTemplateNoParameters() throws NullPointerException, IllegalArgumentException {
		Node subChild = createNode("node2", NO_ATTRIBUTE);
		Node child1 = createNode("node1", NO_ATTRIBUTE, subChild);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child1);
		Template template = createTemplate(templateNode);

		Assert.assertEquals(template.getName(), TEMPLATE_NAME);
		Assert.assertEquals(template.getParameterCount(), 0);
		Assert.assertTrue(template.getRoot().equals(child1));
	}

	@Test
	public void testTemplateMultipleParameters1Child() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", createTemplateParam("param1"), "pos",
				createTemplateParam("param2"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Assert.assertEquals(template.getName(), TEMPLATE_NAME);
		Assert.assertEquals(template.getParameterCount(), 2);
		Assert.assertTrue(template.hasParameter("param1"));
		Assert.assertTrue(template.hasParameter("param2"));
	}

	@Test
	public void testTemplateMultipleParametersMultipleChildren() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet1 = createAttributeSet("text", createTemplateParam("param1"), "pos",
				createTemplateParam("param2"));
		Set<Attribute> attributeSet2 = createAttributeSet("text", createTemplateParam("param3"));
		Node subChild = createNode("node2", attributeSet2);
		Node child1 = createNode("node1", attributeSet1, subChild);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child1);
		Template template = createTemplate(templateNode);

		Assert.assertEquals(template.getName(), TEMPLATE_NAME);
		Assert.assertEquals(template.getParameterCount(), 3);
		Assert.assertTrue(template.hasParameter("param1"));
		Assert.assertTrue(template.hasParameter("param2"));
		Assert.assertTrue(template.hasParameter("param3"));
	}

	@Test
	public void testMultipleParametersIn1Attribute() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text",
				"Hello " + createTemplateParam("firstname") + " " + createTemplateParam("lastname"));
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Assert.assertEquals(template.getName(), TEMPLATE_NAME);
		Assert.assertEquals(template.getParameterCount(), 2); // 2 parameters in one attribute
		Assert.assertTrue(template.hasParameter("firstname"));
		Assert.assertTrue(template.hasParameter("lastname"));
	}

	@Test
	public void testSameParameterInMultipleAttributes1Child() throws NullPointerException, IllegalArgumentException {
		String param1 = createTemplateParam("param1");
		Set<Attribute> attributeSet = createAttributeSet("text", param1, "text2", param1);
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Assert.assertEquals(template.getName(), TEMPLATE_NAME);
		Assert.assertEquals(template.getParameterCount(), 1);
		Assert.assertTrue(template.hasParameter("param1"));
	}

	@Test
	public void testSameParameterInMultipleAttributesMultipleChildren()
			throws NullPointerException, IllegalArgumentException {
		String param1 = createTemplateParam("param1");
		Set<Attribute> attributeSet = createAttributeSet("text", param1);
		Set<Attribute> subChildAttributeSet = createAttributeSet("text", param1);
		Node subChild = createNode("node2", subChildAttributeSet);
		Node child = createNode("node1", attributeSet, subChild);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Assert.assertEquals(template.getName(), TEMPLATE_NAME);
		Assert.assertEquals(template.getParameterCount(), 1);
		Assert.assertTrue(template.hasParameter("param1"));
	}

	// Valid template instantiation

	@Test
	public void testTemplateInstantiationNoParameter() throws NullPointerException, IllegalArgumentException {
		Node child = createNode("node1", NO_ATTRIBUTE);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Node instantiation = createNode(TEMPLATE_NAME, NO_ATTRIBUTE);
		Assert.assertTrue(template.instantiate(instantiation).contains(child));
	}

	@Test
	public void testTemplateInstantiationOneParameter() throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text", "Hi " + createTemplateParam("name") + "!");
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Set<Attribute> instantiationAttributes = createAttributeSet("name", "John");
		Node instantiation = createNode(TEMPLATE_NAME, instantiationAttributes);

		Set<Attribute> expectedAttributes = createAttributeSet("text", "Hi John!");
		Node expectedNode = createNode("node1", expectedAttributes);

		List<Node> instantiatedNode = template.instantiate(instantiation);
		Assert.assertEquals(instantiatedNode.size(), 1);
		Assert.assertTrue(instantiatedNode.contains(expectedNode));
	}

	@Test
	public void testTemplateInstantiationMultipleParametersOneAttribute()
			throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet = createAttributeSet("text",
				"Hi " + createTemplateParam("firstname") + " " + createTemplateParam("lastname") + "!");
		Node child = createNode("node1", attributeSet);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Set<Attribute> instantiationAttributes = createAttributeSet("firstname", "John", "lastname", "Doe");
		Node instantiation = createNode(TEMPLATE_NAME, instantiationAttributes);

		Set<Attribute> expectedAttributes = createAttributeSet("text", "Hi John Doe!");
		Node expectedNode = createNode("node1", expectedAttributes);

		List<Node> instantiatedNode = template.instantiate(instantiation);
		Assert.assertEquals(instantiatedNode.size(), 1);
		Assert.assertTrue(instantiatedNode.contains(expectedNode));
	}

	@Test
	public void testTemplateInstantiationMultipleParametersMultipleChildren()
			throws NullPointerException, IllegalArgumentException {
		Set<Attribute> attributeSet1 = createAttributeSet("text", createTemplateParam("param1"));
		Set<Attribute> attributeSet2 = createAttributeSet("text", createTemplateParam("param2"));
		Node subchild = createNode("node2", attributeSet2);
		Node child = createNode("node1", attributeSet1, subchild);
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Set<Attribute> instantiationAttributes = createAttributeSet("param1", "value1", "param2", "value2");
		Node instantiation = createNode(TEMPLATE_NAME, instantiationAttributes);

		Set<Attribute> expectedAttributes1 = createAttributeSet("text", "value1");
		Set<Attribute> expectedAttributes2 = createAttributeSet("text", "value2");
		Node expectedSubNode = createNode("node2", expectedAttributes2);
		Node expectedNode = createNode("node1", expectedAttributes1, expectedSubNode);

		List<Node> instantiatedNode = template.instantiate(instantiation);
		Assert.assertEquals(instantiatedNode.size(), 1);
		Assert.assertTrue(instantiatedNode.contains(expectedNode));
	}

	@Test
	public void testTemplateInstantationOnlyChildIsContent() {
		Node child = createNode(TEMPLATE_CONTENT_TAG, createAttributeSet(TEMPLATE_CONTENT_ATTR_NAME, "myContent"));
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Node instantationChild = createNode(TEMPLATE_CONTENT_TAG,
				createAttributeSet(TEMPLATE_CONTENT_INSTANTIATION_ATTR_NAME, "myContent"));
		Node instantiation = createNode(TEMPLATE_NAME, NO_ATTRIBUTE, instantationChild);

		List<Node> instantiatedNode = template.instantiate(instantiation);
		Assert.assertEquals(instantiatedNode.size(), 0);
	}

	@Test
	public void testTemplateInstantationNoChildInContent() {
		Node child = createNode("node1", NO_ATTRIBUTE,
				createNode(TEMPLATE_CONTENT_TAG, createAttributeSet(TEMPLATE_CONTENT_ATTR_NAME, "myContent")));
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Node instantationChild = createNode(TEMPLATE_CONTENT_TAG,
				createAttributeSet(TEMPLATE_CONTENT_INSTANTIATION_ATTR_NAME, "myContent"));
		Node instantiation = createNode(TEMPLATE_NAME, NO_ATTRIBUTE, instantationChild);

		List<Node> instantiatedNode = template.instantiate(instantiation);
		Assert.assertEquals(instantiatedNode.size(), 1);
		Assert.assertEquals(instantiatedNode.get(0), createNode("node1", NO_ATTRIBUTE));
	}

	@Test
	public void testTemplateInstantationChildAddedToTemplateInstantiation() {
		String param1 = createTemplateParam("param1");
		Set<Attribute> attributeSet = createAttributeSet("text", param1);
		Node child = createNode("node1", attributeSet,
				createNode(TEMPLATE_CONTENT_TAG, createAttributeSet(TEMPLATE_CONTENT_ATTR_NAME, "myContent")));
		Node templateNode = createTemplateNodeWithDefaultOptionValues(TEMPLATE_NAME, child);
		Template template = createTemplate(templateNode);

		Set<Attribute> instantiationAttributes = createAttributeSet("param1", "value1");
		Node instantationChild = createNode(TEMPLATE_CONTENT_TAG,
				createAttributeSet(TEMPLATE_CONTENT_INSTANTIATION_ATTR_NAME, "myContent"),
				createNode("node2", NO_ATTRIBUTE));
		Node instantiation = createNode(TEMPLATE_NAME, instantiationAttributes, instantationChild);

		Set<Attribute> expectedAttributes = createAttributeSet("text", "value1");
		Node expectedSubNode = createNode("node2", NO_ATTRIBUTE);
		Node expectedNode = createNode("node1", expectedAttributes, expectedSubNode);

		List<Node> instantiatedNode = template.instantiate(instantiation);
		Assert.assertEquals(instantiatedNode.size(), 1);
		Assert.assertTrue(instantiatedNode.contains(expectedNode));
	}
}
