/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2010-2021, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net;

import java.net.UnknownHostException;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.NetworkAddressNatives;
import com.is2t.support.net.util.IPAddressConstants;

import ej.error.Message;
import ej.sni.SNI;

public class NetworkAddress implements INetworkAddress {
	private static final int HOST_NAME_MAX_SIZE = 128;

	@Override
	public String getLocalHostname() throws UnknownHostException{
		byte[] localHostNameBuffer = new byte[HOST_NAME_MAX_SIZE];
		int err = NetworkAddressNatives.getLocalHostnameNative(localHostNameBuffer, localHostNameBuffer.length, false);

		//retry native returned with result
		while(err == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			err = NetworkAddressNatives.getLocalHostnameNative(localHostNameBuffer, localHostNameBuffer.length,
					true);
		}

		if (err < 0) {
			throw new UnknownHostException();
		}
		return SNI.toJavaString(localHostNameBuffer);
	}

	@Override
	public byte[] lookupInaddrAny() throws UnknownHostException {
		byte[] addrBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		int len = NetworkAddressNatives.lookupInaddrAny(addrBuffer, addrBuffer.length, false);

		//retry native returned with result
		while(len == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			len = NetworkAddressNatives.lookupInaddrAny(addrBuffer, addrBuffer.length, true);
		}

		if (len < 0) {
			throw new UnknownHostException(Message.at(NetErrors.getInstance(), len));
		}
		byte[] addr;
		System.arraycopy(addrBuffer, 0, addr = new byte[len], 0, len);
		return addr;
	}

	@Override
	public byte[] loopbackAddress() throws UnknownHostException{
		byte[] addrBuffer = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		int len = NetworkAddressNatives.loopbackAddress(addrBuffer, addrBuffer.length, false);

		//retry native returned with result
		while(len == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			len = NetworkAddressNatives.loopbackAddress(addrBuffer, addrBuffer.length, true);
		}

		if (len < 0) {
			throw new UnknownHostException(Message.at(NetErrors.getInstance(), len));
		}
		byte[] addr;
		System.arraycopy(addrBuffer, 0, addr = new byte[len], 0, len);
		return addr;
	}
}
