/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2014-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;

public class NetworkInterfaceNatives {

	/**
	 * Gets the hardware address associated with the given network interface name {@code ifname}.
	 * <p>When this method returns, the buffer {@code hwAddr} is filled with the hardware address.
	 *
	 * @param ifname the network interface name buffer
	 * @param ifnameLen the network interface name length
	 * @param hwAddr the buffer into which the hardware address will be stored
	 * @param hwAddrMaxLength maximum size of hardware address that can be stored in the buffer {@code hwAddr}
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the hardware address size in bytes, 0 if no hardware address found or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getHardwareAddress(byte[] ifname,int ifnameLen,
			byte[] hwAddr, int hwAddrMaxLength, boolean retry);

	/**
	 * Gets the network interface name at the {@code index} position in the network interfaces list.
	 * When this method returns, the {@link ifname} buffer is filled with the network interface name
	 *
	 * @param index the network interface index
	 * @param ifname the buffer into which the network interface name will be stored
	 * @param length the buffer length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the network interface name length, 0 if no interface found or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterface(int index, byte[] ifname, int length, boolean retry);


	/**
	 * Returns the number of the network interface addresses.
	 *
	 * @param index the network interface index-position in the interfaces list
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the number of the network interface addresses, or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterfaceAddressesCount(int index, boolean retry);


	/**
	 * Gets a network interface address in the interface addresses list.
	 * <p>When this method returns, the {@code addr} buffer is filled with the network interface address.
	 *
	 * @param idxIf the index-position of the interface in the interfaces list
	 * @param idxAddr the index-position of the address in the interface addresses list
	 * @param addr the buffer into which the address will be stored
	 * @param length the buffer length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the network interface address size in bytes or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterfaceAddress(int idxIf, int idxAddr, byte[] addr, int length, boolean retry);

	/**
	 * Returns the number of network interfaces.
	 *
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the number of network interfaces or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getVMInterfacesCount(boolean retry);

	/**
	 * Checks if the given network interface {@code ifname} is a loopback interface.
	 *
	 * @param ifname the network interface name
	 * @param length the network interface name length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 if true, 1 if false or negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int isLoopback(byte[] ifname, int length, boolean retry);

	/**
	 * Checks if the given network interface {@code ifname} is a point-to-point interface.
	 *
	 * @param ifname the network interface name
	 * @param length the network interface name length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 if true, 1 if false or negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int isPointToPoint(byte[] ifname, int length, boolean retry);

	/**
	 * Checks if the given network interface {@code ifname} is up.
	 *
	 * @param ifname the network interface name
	 * @param length the network interface name length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 if true, 1 if false or negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int isUp(byte[] ifname, int length, boolean retry);

	/**
	 * Checks if the given network interface {@code ifname} supports multicast.
	 *
	 * @param name the network interface name
	 * @param length the network interface name length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 if true, 1 if false or negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int supportsMulticast(byte[] ifname, int length, boolean retry);

	/**
	 * Gets the the Maximum Transmission Unit (MTU) of the interface with the
	 * given network interface name {@code ifname}.
	 *
	 * @param ifname
	 *            the network interface name buffer
	 * @param ifnameLen
	 *            the network interface name length
	 * @param retry
	 *            true when the previous call returned
	 *            {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT} and
	 *            the calling process repeats the call to this operation for its
	 *            completion
	 * @return the MTU as an int, or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getMTU(byte[] ifname, int ifnameLen, boolean retry);

}
