/*
 * Java
 *
 * Copyright 2018-2019 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.support.security.securerandom;

import java.security.ProviderException;
import java.security.SecureRandomSpi;

import com.is2t.support.security.NativeResource;

import ej.sni.NativeException;

/**
 *
 */
public class NativeSecureRandomSpi extends SecureRandomSpi {

	/**
	 * Native identifier for the secure random engine. Initialized during engine
	 * init.
	 */
	private final int nativeId;

	/**
	 * Native resource related to the PRNG. Initialized during engine init.
	 */
	private final NativeSecureRandomResource nativeResource;

	/**
	 *
	 */
	public NativeSecureRandomSpi() {
		super();

		int nativeId;
		try {
			// Maybe some resources was garbage collected.
			// Try to release their related underlying native resources without trigger
			// the GC before.
			NativeResource.recycle(false);

			nativeId = nativeInit();

		} catch (NativeException e) {
			// An error occurs during initialization, may be there is no space left to
			// allocate new native resource.
			// Trigger a new fresh GC and release new potential useless resources and try
			// the initialization again!
			NativeResource.recycle(true);
			try {
				nativeId = nativeInit();
			} catch (NativeException e2) {
				// Wrap any native exception into a ProviderException.
				throw new ProviderException(e2);
			}
		}

		// the initialization succeeds; create the associated resource.
		this.nativeId = nativeId;
		this.nativeResource = new NativeSecureRandomResource(this, this.nativeId);

	}

	@Override
	protected void engineSetSeed(byte[] seed) {
		nativeSetSeed(nativeId, seed, seed.length);
	}

	@Override
	protected void engineNextBytes(byte[] bytes) {
		nativeNextBytes(nativeId, bytes, bytes.length);
	}

	@Override
	protected byte[] engineGenerateSeed(int numBytes) {
		byte[] seed = new byte[numBytes];
		nativeGenerateSeed(nativeId, seed, numBytes);
		return seed;
	}

	/**
	 * Initializes the native resource
	 *
	 * @return a nativeId
	 * @throws NativeException
	 */
	native public static int nativeInit() throws NativeException;

	/**
	 * Sets the PRNG seed
	 *
	 * @param nativeId
	 *            the resource id
	 * @param seed
	 *            the seed to use
	 * @throws NativeException
	 */
	native public static void nativeSetSeed(int nativeId, byte[] seed, int seedLength) throws NativeException;

	/**
	 * Generates random bytes
	 *
	 * @param nativeId
	 *            the resource id
	 * @param bytes
	 *            the array to fill with random bytes
	 * @param numBytes
	 *            the number of bytes to generate
	 * @throws NativeException
	 */
	native public static void nativeNextBytes(int nativeId, byte[] bytes, int numBytes) throws NativeException;

	/**
	 * Generates a seed
	 *
	 * @param nativeId
	 *            the resource id
	 * @param bytes
	 *            the seed array to fill with random bytes
	 * @param numBytes
	 *            the seed size in bytes
	 * @throws NativeException
	 */
	native public static void nativeGenerateSeed(int nativeId, byte[] bytes, int numBytes) throws NativeException;

	/**
	 * Closes the native resource
	 *
	 * @param nativeId
	 *            the resource id
	 * @throws NativeException
	 */
	native public static void nativeClose(int nativeId) throws NativeException;
}
