/*
 * Copyright (c) 2003, 2017, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2017-2018 IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.signature;

import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.ProviderException;
import java.security.PublicKey;
import java.security.SignatureException;
import java.security.SignatureSpi;

/**
 * Implementation copied from <code>sun.security.ec.ECDSASignature</code> and
 * <code>sun.security.rsa.RSASignature</code>.
 */
public abstract class AbstractSignatureSpi extends SignatureSpi {

	// message digest implementation we use
	private final MessageDigest messageDigest;

	// flag indicating whether the digest has been reset
	private boolean needsReset;


	public AbstractSignatureSpi(String messageDigestAlgorithmName) {
		try {
			messageDigest = MessageDigest.getInstance(messageDigestAlgorithmName);
		} catch (NoSuchAlgorithmException e) {
			throw new ProviderException(e);
		}
		//needsReset = false; default
	}

	@Override
	protected void engineInitVerify(PublicKey publicKey) throws InvalidKeyException {
		setKeys(publicKey, null);
		resetDigest();
	}

	@Override
	protected void engineInitSign(PrivateKey privateKey) throws InvalidKeyException {
		setKeys(null, privateKey);
		resetDigest();
	}

	/**
	 * Resets the message digest if needed.
	 */
	protected void resetDigest() {
		if (needsReset) {
			if (messageDigest != null) {
				messageDigest.reset();
			}
			needsReset = false;
		}
	}

	/**
	 * Returns the message digest value.
	 */
	protected byte[] getDigestValue() throws SignatureException {
		needsReset = false;
		return messageDigest.digest();
	}

	// update the signature with the plaintext data. See JCA doc
	@Override
	protected void engineUpdate(byte b) throws SignatureException {
		messageDigest.update(b);
		needsReset = true;
	}

	// update the signature with the plaintext data. See JCA doc
	@Override
	protected void engineUpdate(byte[] b, int off, int len) throws SignatureException {
		messageDigest.update(b, off, len);
		needsReset = true;

	}

	@Override
	protected byte[] engineSign() throws SignatureException {

		return implSign(getDigestValue());
	}

	@Override
	protected boolean engineVerify(byte[] sigBytes) throws SignatureException {

		return implVerify(sigBytes, getDigestValue());
	}

	abstract protected boolean implVerify(byte[] sigBytes, byte[] digestValue);

	abstract protected byte[] implSign(byte[] digestValue);

	/**
	 * Sets the public and private keys for this signature.
	 * @param publicKey the public key to set
	 * @param privateKey the private key to set
	 * @throws InvalidKeyException if one of the keys is invalid.
	 */
	abstract protected void setKeys(PublicKey publicKey, PrivateKey privateKey) throws InvalidKeyException;

}
