/*
 * Copyright (c) 2009, 2017, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2017-2018 IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.signature;

import java.security.InvalidKeyException;
import java.security.PrivateKey;
import java.security.PublicKey;

import com.is2t.support.security.NativePrivateKey;
import com.is2t.support.security.NativePublicKey;

import ej.sni.NativeException;

/**
 *
 */
public class NativeSignatureSpi extends AbstractSignatureSpi {

	private final int nativeAlgorithmId;

	private NativePublicKey publicKey;
	private NativePrivateKey privateKey;

	public NativeSignatureSpi(int nativeAlgorithmId, String messageDigestAlgorithmName){
		super(messageDigestAlgorithmName);
		this.nativeAlgorithmId = nativeAlgorithmId;
	}

	@Override
	protected boolean implVerify(byte[] sigBytes, byte[] digestValue) {

		byte[] nativeKeyData = publicKey.getData();

		return nativeVerify(nativeAlgorithmId, sigBytes, sigBytes.length, nativeKeyData, nativeKeyData.length, digestValue, digestValue.length);
	}

	@Override
	protected byte[] implSign(byte[] digestValue) {

		byte[] nativeKeyData = privateKey.getData();

		byte[] sigBytes = new byte[nativeKeyData.length];

		int sigLen = nativeSign(nativeAlgorithmId, sigBytes, sigBytes.length, nativeKeyData, nativeKeyData.length,
				digestValue, digestValue.length);
		byte[] signature = new byte[sigLen];
		System.arraycopy(sigBytes, 0, signature, 0, sigLen);
		return signature;
	}

	@Override
	protected void setKeys(PublicKey publicKey, PrivateKey privateKey) throws InvalidKeyException {
		try {
			this.publicKey = (NativePublicKey) publicKey;
			this.privateKey = (NativePrivateKey) privateKey;
		}
		catch(ClassCastException e) {
			throw new InvalidKeyException(e);
		}
	}

	/**
	 * Gets for the given algorithm the signature algorithm description.
	 * <p>
	 *
	 * @param algorithm Null terminated string that describes the algorithm
	 * @return algorithm ID on success or -1 on error.
	 */
	native public static int nativeGetAlgorithmDescription(byte[] algorithm, byte[] digestAlgorithmName, int digestAlgorithmNameLength);

	/**
	 * Verifies the given signature.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 * @param signature
	 *            the signature to be verified
	 * @param signatureLength
	 *            the length of the signature
	 * @param keyData
	 *            the public key data
	 * @param keyDataLength
	 *            the length of the public key data
	 * @param digest
	 *            the digest of the data to be checked against the signature
	 * @param digestLength
	 *            the length of the digest
	 * @return true if the signature was verified, false if not
	 * @throws NativeException
	 */
	native private static boolean nativeVerify(int nativeAlgorithmId, byte[] signature, int signatureLength,
			byte[] keyData, int keyDataLength, byte[] digest, int digestLength) throws NativeException;

	/**
	 * Signs the given data.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 * @param signature
	 *            the array that will contain the resulting signature
	 * @param signatureLength
	 *            the length of the signature
	 * @param keyData
	 *            the private key data
	 * @param keyDataLength
	 *            the length of the private key data
	 * @param digest
	 *            the digest of the data to be signed
	 * @param digestLength
	 *            the length of the digest
	 * @throws NativeException
	 */
	native private static int nativeSign(int nativeAlgorithmId, byte[] signature, int signatureLength,
			byte[] keyData, int keyDataLength, byte[] digest, int digestLength) throws NativeException;
}
