/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.vivante.utils;

import com.verisilicon.vivante.vg_lite;
import com.verisilicon.vivante.vg_lite_gradient;
import com.verisilicon.vivante.vg_lite_matrix;
import com.verisilicon.vivante.vg_lite_path;

import ej.microui.display.GraphicsContext;

/**
 * Represents a path with a gradient filling.
 * <p>
 * The gradient is created dynamically before each draw (used by immutables generator).
 */
public class DynamicGradientPath implements Path {
	private final vg_lite_path path;
	private final float x1;
	private final float y1;
	private final float length;
	private final float angle;
	private final int[] colors;
	private final int[] stops;
	private final int fillRule;

	/**
	 * Creates a colored path.
	 *
	 * @param path
	 *            the path.
	 * @param x1
	 *            the x1.
	 * @param y1
	 *            the y1.
	 * @param length
	 *            the length.
	 * @param angle
	 *            the angle.
	 * @param colors
	 *            the colors.
	 * @param stops
	 *            the stops.
	 * @param fillRule
	 *            the fill rule (one of {@link vg_lite#VG_LITE_FILL_EVEN_ODD} and
	 *            {@link vg_lite#VG_LITE_FILL_NON_ZERO}).
	 */
	public DynamicGradientPath(vg_lite_path path, float x1, float y1, float length, float angle, int[] colors,
			int[] stops, int fillRule) {
		this.path = path;
		this.x1 = x1;
		this.y1 = y1;
		this.length = length;
		this.angle = angle;
		this.colors = colors;
		this.stops = stops;
		this.fillRule = fillRule;
	}

	@Override
	public vg_lite_path getVGLitePath() {
		return this.path;
	}

	@Override
	public void draw(GraphicsContext g, vg_lite_matrix matrix) {
		vg_lite_gradient gradient = new vg_lite_gradient();
		vg_lite_matrix gradientMatrix = gradient.getMatrix();
		vg_lite_matrix.vg_lite_identity(gradientMatrix);
		vg_lite_matrix.vg_lite_translate(this.x1, this.y1, gradientMatrix); // third transformation
		vg_lite_matrix.vg_lite_rotate(this.angle, gradientMatrix); // second transformation
		vg_lite_matrix.vg_lite_scale(this.length / 256, 1, gradientMatrix); // first transformation
		vg_lite_gradient.vg_lite_set_grad(gradient, this.colors.length, this.colors, this.stops);
		vg_lite.vg_lite_draw_gradient(g, this.path, this.fillRule, matrix, gradient, vg_lite.VG_LITE_BLEND_SRC_OVER);
		gradient.close();
	}
}
