/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.vivante.utils;

import static com.verisilicon.vivante.vg_lite_path.VG_LITE_S8;

import com.verisilicon.vivante.vg_lite_path;

/**
 * Utility class for building {@link vg_lite_path} in S8 format.
 *
 * <p>
 * See Section <code>2.4 Vector Path Data Opcodes</code> of <code>VGLite® Graphics API v2.07</code>.
 * </p>
 */
public class PathBufferS8 extends PathBufferBase {
	private static final int COORD_SIZE = 1;

	private byte xOffset;

	private byte yOffset;

	@Override
	protected int getUnitSize() {
		return COORD_SIZE;
	}

	@Override
	public int getFormat() {
		return VG_LITE_S8;
	}

	@Override
	protected void command(int cmd) {
		byte[] path = extend(1);
		int index = this.index;

		path[index++] = (byte) cmd;
		this.index = index;
	}

	/**
	 * Adds a command with 1 point parameter in the path data
	 *
	 * @param cmd
	 *            the command to add in the path data
	 * @param x
	 *            the horizontal coordinate of the point parameter
	 * @param y
	 *            the vertical coordinate of the point parameter
	 */
	protected void command(int cmd, byte x, byte y) {
		byte[] path = extend(3);
		int index = this.index;

		path[index++] = (byte) cmd;
		path[index++] = (x);
		path[index++] = (y);
		this.index = index;
	}

	/**
	 * Adds a command with 2 points parameter in the path data
	 *
	 * @param cmd
	 *            the command to add in the path data
	 * @param x1
	 *            the horizontal coordinate of the first point parameter
	 * @param y1
	 *            the vertical coordinate of the first point parameter
	 * @param x2
	 *            the horizontal coordinate of the second point parameter
	 * @param y2
	 *            the vertical coordinate of the second point parameter
	 */
	protected void command(int cmd, byte x1, byte y1, byte x2, byte y2) {
		byte[] path = extend(5);
		int index = this.index;

		path[index++] = (byte) cmd;
		path[index++] = (x1);
		path[index++] = (y1);
		path[index++] = (x2);
		path[index++] = (y2);
		this.index = index;
	}

	/**
	 * Adds a command with 3 points parameter in the path data
	 *
	 * @param cmd
	 *            the command to add in the path data
	 * @param x1
	 *            the horizontal coordinate of the first point parameter
	 * @param y1
	 *            the vertical coordinate of the first point parameter
	 * @param x2
	 *            the horizontal coordinate of the second point parameter
	 * @param y2
	 *            the vertical coordinate of the second point parameter
	 * @param x3
	 *            the horizontal coordinate of the third point parameter
	 * @param y3
	 *            the vertical coordinate of the third point parameter
	 */
	protected void command(int cmd, byte x1, byte y1, byte x2, byte y2, byte x3, byte y3) {
		byte[] path = extend(7);
		int index = this.index;

		path[index++] = (byte) cmd;
		path[index++] = (x1);
		path[index++] = (y1);
		path[index++] = (x2);
		path[index++] = (y2);
		path[index++] = (x3);
		path[index++] = (y3);
		this.index = index;
	}

	/**
	 * Sets an offset that will be used during drawing.
	 *
	 * @param x
	 *            The horizontal offset
	 * @param y
	 *            The vertical offset
	 */
	public void setOffset(byte x, byte y) {
		this.xOffset = x;
		this.yOffset = y;
	}

	/**
	 * Applies the offset to an horizontal coordinate
	 *
	 * Note: This method is duplicated in <code>PathBufferS8</code>, <code>PathBufferS16</code>,
	 * <code>PathBufferS32</code> and <code>PathBufferFP32</code>.
	 *
	 * @param x
	 *            The horizontal coordinate to offset
	 * @return the updated coordinate
	 */
	private byte applyXOffset(byte x) {
		return (byte) (x + this.xOffset);
	}

	/**
	 * Applies the offset to an vertical coordinate
	 *
	 * Note: This method is duplicated in <code>PathBufferS8</code>, <code>PathBufferS16</code>,
	 * <code>PathBufferS32</code> and <code>PathBufferFP32</code>.
	 *
	 * @param y
	 *            The vertical coordinate to offset
	 * @return the updated coordinate
	 */
	private byte applyYOffset(byte y) {
		return (byte) (y + this.yOffset);
	}

	/**
	 * Add a move command to the given point.
	 *
	 * The current point is set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param x
	 *            the horizontal coordinate of the point to move to
	 * @param y
	 *            the vertical coordinate of the point to move to
	 */
	public void move(byte x, byte y) {
		command(CMD_MOVE, applyXOffset(x), applyYOffset(y));
	}

	/**
	 * Add a move command to the given point.
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param x
	 *            the horizontal coordinate of the point to move to
	 * @param y
	 *            the vertical coordinate of the point to move to
	 */
	public void moveRel(byte x, byte y) {
		command(CMD_MOVE, x, y);
	}

	/**
	 * Add a line command from the current point to the given point
	 *
	 * A line is drawn from a starting point (current point) to an ending point (c, y).
	 *
	 * The current point is then set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param x
	 *            the horizontal coordinate of the end of the line
	 * @param y
	 *            the vertical coordinate of the end of the line
	 */
	public void line(byte x, byte y) {
		x = applyXOffset(x);
		y = applyYOffset(y);

		command(CMD_LINE, x, y);
	}

	/**
	 * Add a line command from the current point to the given point
	 *
	 * A line is drawn from a starting point (current point) to an ending point (c, y).
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param x
	 *            the horizontal coordinate of the end of the line relative to the current point
	 * @param y
	 *            the vertical coordinate of the end of the line relative to the current point
	 */
	public void lineRel(byte x, byte y) {
		command(CMD_LINE_REL, x, y);
	}

	/**
	 * Add a quadratic curve from the current point to the given point
	 *
	 * A quadratic curve is drawn from a starting point (current point) to an ending point (x, y) using one control
	 * point (cx, cy).
	 *
	 * The current point is then set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param cx
	 *            the horizontal coordinate of the control point
	 * @param cy
	 *            the vertical coordinate of the control point
	 * @param x
	 *            the horizontal coordinate of the end of the curve
	 * @param y
	 *            the vertical coordinate of the end of the curve
	 */
	public void quad(byte cx, byte cy, byte x, byte y) {
		x = applyXOffset(x);
		y = applyYOffset(y);

		command(CMD_QUAD, cx, cy, x, y);
	}

	/**
	 * Add a quadratic curve from the current point to the given point
	 *
	 * A quadratic curve is drawn from a starting point (current point) to an ending point (x, y) using one control
	 * point (cx, cy).
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param cx
	 *            the horizontal coordinate of the control point relative to the current point
	 * @param cy
	 *            the vertical coordinate of the control point relative to the current point
	 * @param x
	 *            the horizontal coordinate of the end of the curve relative to the current point
	 * @param y
	 *            the vertical coordinate of the end of the curve relative to the current point
	 */
	public void quadRel(byte cx, byte cy, byte x, byte y) {
		command(CMD_QUAD_REL, cx, cy, x, y);
	}

	/**
	 * Add a cubic curve from the current point to the given point
	 *
	 * A cubic curve is drawn from a starting point (current point) to an ending point (x, y) using two control point
	 * (cx1, cy1) and (cx2, cy2).
	 *
	 * The current point is then set to (x, y).
	 *
	 * All point coordinates are absolute.
	 *
	 * @param cx1
	 *            the horizontal coordinate of the first control point
	 * @param cy1
	 *            the vertical coordinate of the first control point
	 * @param cx2
	 *            the horizontal coordinate of the second control point
	 * @param cy2
	 *            the vertical coordinate of the second control point
	 * @param x
	 *            the horizontal coordinate of the end of the curve
	 * @param y
	 *            the vertical coordinate of the end of the curve
	 */
	public void cubic(byte cx1, byte cy1, byte cx2, byte cy2, byte x, byte y) {
		x = applyXOffset(x);
		y = applyYOffset(y);

		command(CMD_CUBIC, cx1, cy1, cx2, cy2, x, y);
	}

	/**
	 * Add a cubic curve from the current point to the given point
	 *
	 * A cubic curve is drawn from a starting point (current point) to an ending point (x, y) using two control point
	 * (cx1, cy1) and (cx2, cy2).
	 *
	 * The current point is set to (current_x + x, current_y + y).
	 *
	 * All point coordinates are relative to the current point.
	 *
	 * @param cx1
	 *            the horizontal coordinate of the first control point
	 * @param cy1
	 *            the vertical coordinate of the first control point
	 * @param cx2
	 *            the horizontal coordinate of the second control point
	 * @param cy2
	 *            the vertical coordinate of the second control point
	 * @param x
	 *            the horizontal coordinate of the end of the curve
	 * @param y
	 *            the vertical coordinate of the end of the curve
	 */
	public void cubicRel(byte cx1, byte cy1, byte cx2, byte cy2, byte x, byte y) {
		command(CMD_CUBIC_REL, cx1, cy1, cx2, cy2, x, y);
	}
}
