/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.vivante.utils;

import java.util.NoSuchElementException;

import com.verisilicon.vivante.vg_lite;
import com.verisilicon.vivante.vg_lite_matrix;

import ej.bon.Immutables;
import ej.microui.display.GraphicsContext;

/**
 * Represents an SVG image (converted from a .svg image with svg2mvg tool).
 */
public class SvgImage {

	private final Path[] paths;

	/**
	 * Creates a SVG image with its paths.
	 *
	 * @param paths
	 *            the paths of the SVG image.
	 */
	public SvgImage(Path[] paths) {
		this.paths = paths;
	}

	/**
	 * Gets the SVG image instance with the given name.
	 *
	 * @param name
	 *            the name of the image (example: <code>mascot_microej.svg</code>).
	 * @return the SVG image.
	 * @throws NoSuchElementException
	 *             if the image does not exist.
	 */
	public static SvgImage getSvgImage(String name) {
		try {
			return (SvgImage) Immutables.get(name);
		} catch (ClassCastException e) {
			throw new NoSuchElementException(e.toString());
		}
	}

	/**
	 * Draws the paths with an identity matrix, {@link vg_lite#VG_LITE_FILL_EVEN_ODD} and
	 * {@link vg_lite#VG_LITE_BLEND_NONE}.
	 * <p>
	 * Warning: this method is compatible with com.microej.api#vivante-vglite-0.5.0 or later (new matrix is an identity
	 * matrix).
	 *
	 * @param g
	 *            the graphics context to draw on.
	 */
	public void draw(GraphicsContext g) {
		vg_lite_matrix matrix = new vg_lite_matrix();
		draw(g, matrix);
	}

	/**
	 * Draws the paths with the given parameters.
	 *
	 * @param g
	 *            the graphics context to draw on.
	 * @param matrix
	 *            the matrix to apply on the paths.
	 */
	public void draw(GraphicsContext g, vg_lite_matrix matrix) {
		for (Path path : this.paths) {
			path.draw(g, matrix);
		}
	}
}
