/*
 * Java
 *
 * Copyright 2016-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.ecom.network;

import java.io.IOException;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import ej.bon.Immortals;
import ej.ecom.network.natives.InterfaceManagerImplNatives;
import ej.ecom.network.natives.InterfaceManagerImplNativesConstants;
import ej.sni.SNI;

/**
 * A manager for the network interfaces.
 *
 * It allows to get the current IP configuration of an interface and to reconfigure it.
 *
 * It allows to enable/disable and start/stop an interface. Enable/disable is simply a logical state of the interface
 * managed by software. Start/stop is more hardware-related. It may physically turn on/off the hardware module and it is
 * intended to manage power consumption. Scenario example: a system uses Ethernet as its main interface and 3G as its
 * backup interface. In normal mode, Ethernet is started and enabled while 3G is stopped (and hence disabled) and
 * communications all go through the Ethernet interface. When Ethernet loses Internet connection, it is disabled and 3G
 * is started. Communications will then go through the 3G interface. When Ethernet gets Internet connection back, it is
 * re-enabled, 3G is stopped and the system goes back to normal mode.
 */
public class NetworkInterfaceManager {

	private static final int J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT = -22;

	private static byte[] netifName;
	private static byte[] address;
	private static Object monitor = new Object();

	static {
		Immortals.run(new Runnable() {
			@Override
			public void run() {
				netifName = new byte[64]; // arbitrary size
				address = new byte[4]; // 4 bytes in ipv4
			}
		});
	}

	public static void enable(NetworkInterface netif) throws IllegalStateException, IOException {
		if (netif == null) {
			throw new IllegalArgumentException();
		}

		synchronized (monitor) {
			SNI.toCString(netif.getName(), netifName);

			int err = InterfaceManagerImplNatives.enable(netifName, 0,
					netifName.length, false);

			// retry native returned with result
			while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				err = InterfaceManagerImplNatives.enable(netifName, 0,
						netifName.length, true);
			}

			if (err < 0) {
				throw new IllegalStateException();
			}

		}
	}

	public static void disable(NetworkInterface netif) throws IllegalStateException, IOException {
		if (netif == null) {
			throw new IllegalArgumentException();
		}

		synchronized (monitor) {
			SNI.toCString(netif.getName(), netifName);

			int err = InterfaceManagerImplNatives.disable(netifName, 0,
					netifName.length, false);

			// retry native returned with result
			while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				err = InterfaceManagerImplNatives.disable(netifName, 0,
						netifName.length, true);
			}

			if (err < 0) {
				throw new IllegalStateException();
			}
		}

	}

	public static boolean isEnabled(NetworkInterface netif) {
		if (netif == null) {
			throw new IllegalArgumentException();
		}

		synchronized (monitor) {
			SNI.toCString(netif.getName(), netifName);

			int res = InterfaceManagerImplNatives.isEnabled(netifName, 0, netifName.length, false);

			// retry native returned with result
			while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				res = InterfaceManagerImplNatives.isEnabled(netifName, 0,
						netifName.length, true);
			}

			// error case
			if (res < 0) {
				return false;
			}

			if (res == InterfaceManagerImplNativesConstants.IF_STATE_ENABLED) {
				return true;
			}

			return false;
		}
	}

	public static void start(NetworkInterface netif) throws IOException {
		if (netif == null) {
			throw new IllegalArgumentException();
		}

		synchronized (monitor) {
			SNI.toCString(netif.getName(), netifName);

			int err = InterfaceManagerImplNatives.start(netifName, 0,
					netifName.length,
					false);

			// retry native returned with result
			while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				err = InterfaceManagerImplNatives.start(netifName, 0,
						netifName.length, true);
			}

			// error case
			if (err < 0) {
				throw new IOException("Cannot start interface (err="+err+")");
			}

		}
	}

	public static void stop(NetworkInterface netif) throws IOException {
		synchronized (monitor) {
			// First, disable the interface
			disable(netif);

			// Then, stop the interface
			SNI.toCString(netif.getName(), netifName);

			int err = InterfaceManagerImplNatives.stop(netifName, 0,
					netifName.length,
					false);

			// retry native returned with result
			while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				err = InterfaceManagerImplNatives.stop(netifName, 0,
						netifName.length, true);
			}

			// error case
			if (err < 0) {
				throw new IOException("Cannot stop interface (err="+err+")");
			}

		}
	}

	public static boolean isStarted(NetworkInterface netif) {
		if (netif == null) {
			throw new IllegalArgumentException();
		}

		synchronized (monitor) {
			SNI.toCString(netif.getName(), netifName);

			int res = InterfaceManagerImplNatives.isStarted(netifName, 0,
					netifName.length, false);

			// retry native returned with result
			while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				res = InterfaceManagerImplNatives.isStarted(netifName, 0,
						netifName.length, true);
			}

			// error case
			if (res < 0) {
				return false;
			}

			if (res == InterfaceManagerImplNativesConstants.IF_STATE_STARTED) {
				return true;
			}

			return false;
		}
	}

	public static IPConfiguration getIPConfiguration(NetworkInterface netif) throws IOException {
		if (netif == null) {
			throw new IllegalArgumentException();
		}

		synchronized (monitor) {
			SNI.toCString(netif.getName(), netifName);

			IPConfiguration conf = new IPConfiguration();

			// Get whether this is a static configuration or not
			boolean isStatic = isStatic();
			conf.useDHCP(!isStatic);

			// Get Network Interface IP address refresh network interfaces data
			// make a call to NetworkInterface.getNetworkInterfaces() method to
			// update this values by making some natives call
			Enumeration<NetworkInterface> netifs = NetworkInterface.getNetworkInterfaces();
			NetworkInterface tmpNetif =  null;
			while(netifs.hasMoreElements()){
				tmpNetif = netifs.nextElement();
				if((tmpNetif != null) && (tmpNetif.getName().equals(netif.getName()))){
					Enumeration<InetAddress> netifAddrs = tmpNetif.getInetAddresses();
					if(netifAddrs.hasMoreElements()){
						// Network Interface found, set this IP address in the IPConfiguration object
						conf.setIP(netifAddrs.nextElement());
					}
				}
			}

			// Get netmask
			InetAddress netmask = getNetmask();
			conf.setNetmask(netmask);

			// Get gateway
			InetAddress gateway = getGateway();
			conf.setGateway(gateway);

			// Get whether DNS is statically configured or not
			boolean isDNSStatic = isDNSStatic();
			conf.useStaticDNS(isDNSStatic);

			// Get DNS (one by one)
			int count = getDNSCount();
			List<InetAddress> list = new ArrayList<>(count);

			for (int i = 0; i < count; i++) {
				InetAddress dns = getDNS(i);
				if (dns != null) {
					list.add(dns);
				}
			}

			// Convert to array of appropriate size
			InetAddress[] dns = new InetAddress[list.size()];
			list.toArray(dns);
			conf.setDNS(dns);

			return conf;
		}
	}

	public static void configure(NetworkInterface netif, IPConfiguration conf)
			throws IOException {
		if ((netif == null) || (conf == null)) {
			throw new IllegalArgumentException();
		}

		synchronized (monitor) {
			SNI.toCString(netif.getName(), netifName);

			// Tell underlying module to enter configuration mode
			startConfiguration();

			if (conf.isStatic()) {
				// DHCP is not used
				useDHCP(false);

				// IP parameters must be given to the native world
				if (conf.getIP() != null) {
					setIP(conf.getIP());
				}

				if (conf.getNetmask() != null) {
					setNetmask(conf.getNetmask());
				}

				if (conf.getGateway() != null) {
					setGateway(conf.getGateway());
				}

				// Set DNS
				InetAddress[] dns = conf.getDNS();
				if (dns != null) {
					for (int i = 0; i < dns.length; i++) {
						setDNS(dns[i], i);
					}
				} // else: no DNS server was set

			} else {
				// DHCP is used
				useDHCP(true);

				if (conf.isDNSStatic()) {
					// DNS is not provided by DHCP, it is statically configured
					useStaticDNS(true);

					// Set DNS
					InetAddress[] dns = conf.getDNS();
					if (dns != null) {
						for (int i = 0; i < dns.length; i++) {
							setDNS(dns[i], i);
						}
					} // else: no DNS server was set

				} else {
					// DHCP will provide DNS
					useStaticDNS(false);
				}
			}

			// Apply previously set parameters
			endConfiguration();
		}
	}

	private static boolean isStatic() throws IOException {
		int res = InterfaceManagerImplNatives.isStatic(netifName, 0,
				netifName.length, false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = InterfaceManagerImplNatives.isStatic(netifName, 0,
					netifName.length, true);
		}

		// error case
		if (res < 0) {
			throw new IOException("Cannot retrieve IP config (err="+res+")");
		}

		// else, else: native error, consider DHCP (purely arbitrary)

		return res == InterfaceManagerImplNativesConstants.IF_IP_STATIC;
	}

	private static InetAddress getNetmask() throws IOException {
		InetAddress inet = null;

		int err = InterfaceManagerImplNatives.getNetmask(netifName, 0,
				netifName.length, address, 0, address.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.getNetmask(netifName, 0,
					netifName.length, address, 0, address.length, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot retrieve IP config netmask (err="+err+")");
		}

		try {
			StringBuilder sb = new StringBuilder();
			sb.append(String.valueOf((0xFF & address[0])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[1])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[2])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[3])));
			inet = InetAddress.getByName(sb.toString());
		} catch (IllegalArgumentException | UnknownHostException e) {
			// Consider IP is not valid
		}


		return inet;
	}

	private static InetAddress getGateway() throws IOException {
		InetAddress inet = null;

		int err = InterfaceManagerImplNatives.getGateway(netifName, 0,
				netifName.length, address, 0, address.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.getGateway(netifName, 0,
					netifName.length, address, 0, address.length, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot retrieve IP config gateway (err="+err+")");
		}

		try {
			StringBuilder sb = new StringBuilder();
			sb.append(String.valueOf((0xFF & address[0])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[1])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[2])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[3])));
			inet = InetAddress.getByName(sb.toString());
		} catch (IllegalArgumentException | UnknownHostException e) {
			// Consider IP is not valid
		}

		return inet;
	}

	private static boolean isDNSStatic() throws IOException {
		int res = InterfaceManagerImplNatives.isDNSStatic(netifName, 0,
				netifName.length, false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = InterfaceManagerImplNatives.isDNSStatic(netifName, 0,
					netifName.length, true);
		}

		// error case
		if (res < 0) {
			throw new IOException("Cannot retrieve DNS config (err="+res+")");
		}

		return res == InterfaceManagerImplNativesConstants.IF_DNS_STATIC;
	}

	private static int getDNSCount() throws IOException {
		int res = InterfaceManagerImplNatives.getDNSCount(netifName, 0, netifName.length, false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = InterfaceManagerImplNatives.getDNSCount(netifName, 0,
					netifName.length, true);
		}

		// error case
		if (res < 0) {
			throw new IOException("Cannot retrieve the number of registered DNS (err="+res+")");
		}

		return res;
	}

	private static InetAddress getDNS(int index) throws IOException {
		InetAddress inet = null;

		int err = InterfaceManagerImplNatives.getDNS(netifName, 0,
				netifName.length, address, 0, address.length, index, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.getDNS(netifName, 0,
					netifName.length, address, 0, address.length, index, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot retrieve DNS IP address (err="+err+")");
		}

		try {
			StringBuilder sb = new StringBuilder();
			sb.append(String.valueOf((0xFF & address[0])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[1])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[2])));
			sb.append('.');
			sb.append(String.valueOf((0xFF & address[3])));
			inet = InetAddress.getByName(sb.toString());
		} catch (IllegalArgumentException | UnknownHostException e) {
			// Consider IP is not valid
		}

		// else, else: error from native, consider IP is not valid

		return inet;
	}

	private static void startConfiguration() throws IOException {

		int err = InterfaceManagerImplNatives.startConfiguration(netifName, 0,
				netifName.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.startConfiguration(netifName, 0,
					netifName.length, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot apply (start) configuration (err="+err+")");
		}
	}

	private static void useDHCP(boolean use) throws IOException {

		int err = InterfaceManagerImplNatives.useDHCP(netifName, 0, netifName.length,
				use, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.useDHCP(netifName, 0, netifName.length,
					use, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot use DHCP (err="+err+")");
		}
	}

	private static void setIP(InetAddress ip) throws IOException {
		byte[] bytes = ip.getAddress();
		System.arraycopy(bytes, 0, address, 0, address.length);

		int err = InterfaceManagerImplNatives.setIP(netifName, 0,
				netifName.length, address, 0, address.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.setIP(netifName, 0,
					netifName.length, address, 0, address.length, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot set IP (err="+err+")");
		}
	}

	private static void setNetmask(InetAddress netmask) throws IOException {
		byte[] bytes = netmask.getAddress();
		System.arraycopy(bytes, 0, address, 0, address.length);

		int err = InterfaceManagerImplNatives.setNetmask(netifName, 0,
				netifName.length, address, 0, address.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.setNetmask(netifName, 0,
					netifName.length, address, 0, address.length, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot set netmask (err="+err+")");
		}

	}

	private static void setGateway(InetAddress gateway) throws IOException {
		byte[] bytes = gateway.getAddress();
		System.arraycopy(bytes, 0, address, 0, address.length);

		int err = InterfaceManagerImplNatives.setGateway(netifName, 0,
				netifName.length, address, 0, address.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.setGateway(netifName, 0,
					netifName.length, address, 0, address.length, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot set gateway (err="+err+")");
		}
	}

	private static void useStaticDNS(boolean use) throws IOException {

		int err = InterfaceManagerImplNatives.useStaticDNS(netifName, 0,
				netifName.length, use, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.useStaticDNS(netifName, 0,
					netifName.length, use, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot use static DNS (err="+err+")");
		}
	}

	private static void setDNS(InetAddress dns, int index) throws IOException {
		byte[] bytes = dns.getAddress();
		System.arraycopy(bytes, 0, address, 0, address.length);

		int err = InterfaceManagerImplNatives.setDNS(netifName, 0,
				netifName.length, address, 0, address.length, index, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.setDNS(netifName, 0,
					netifName.length, address, 0, address.length, index, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot set DNS (err="+err+")");
		}
	}

	private static void endConfiguration() throws IOException {

		int err = InterfaceManagerImplNatives.endConfiguration(netifName, 0,
				netifName.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = InterfaceManagerImplNatives.endConfiguration(netifName, 0,
					netifName.length, true);
		}

		// error case
		if (err < 0) {
			throw new IOException("Cannot apply (end) configuration (err="+err+")");
		}

	}

}
