/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.ecom.wifi;

/**
 * Wireless Access Point (WAP, often shorten as AP).
 */
public class AccessPoint {

	private String ssid;
	private byte[] bssid;
	private float rssi;
	private int channel;
	private SecurityMode securityMode;
	private WPSMode[] wpsModes;

	/**
	 * Construct a Wifi Access Point instance empty
	 */
	public AccessPoint() {

	}

	/**
	 * Construct a Wifi Access Point instance
	 *
	 * @param ssid
	 *            the Access Point SSID
	 * @param securityMode
	 *            the Access Point security mode
	 */
	public AccessPoint(String ssid, SecurityMode securityMode){
		this.ssid = ssid;
		this.securityMode = securityMode;
	}

	/**
	 * Construct a Wifi Access Point instance
	 * @param ssid the Access Point SSID
	 * @param bssid the Access Point BSSID
	 * @param rssi the Access Point RSSI
	 * @param channel the Access Point channel
	 * @param securityMode the Access Point security mode
	 * @param wpsModes the Access Point WPS capabilities
	 */
	public AccessPoint(String ssid, byte[] bssid, float rssi, int channel, SecurityMode securityMode, WPSMode[] wpsModes) {
		this.ssid = ssid;
		this.bssid = bssid;
		this.rssi = rssi;
		this.channel = channel;
		this.securityMode = securityMode;
		this.wpsModes = wpsModes;
	}

	/**
	 * Get the Service Set IDentification (SSID) of the network.
	 *
	 * @return the SSID
	 */
	public String getSSID() {
		return ssid;
	}

	/**
	 * Set the SSID of the access point.
	 *
	 * @param SSID the SSID
	 */
	public void setSSID(String SSID) {
		this.ssid = SSID;
	}

	/**
	 * Get the Basic Service Set IDentification (BSSID). This is a MAC address of the access point in infrastructure
	 * mode.
	 *
	 * @return the BSSID (consists of 24 + 24 bits = 6 bytes)
	 */
	public byte[] getBSSID() {
		return bssid;
	}

	/**
	 * Set the BSSID of the access point.
	 *
	 * @param BSSID
	 *            the BSSID (must not be null)
	 */
	public void setBSSID(byte[] BSSID) {
		if (BSSID == null) {
			throw new IllegalArgumentException("BSSID must not be null");
		}
		this.bssid = BSSID;
	}

	/**
	 * Get the Received Signal Strength Identification (RSSI) of this access point.
	 *
	 * This is a negative value expressed in dBm.
	 *
	 * @return the RSSI
	 */
	public float getRSSI() {
		return rssi;
	}

	/**
	 * Set the RSSI of the access point.
	 *
	 * @param RSSI
	 *            the RSSI
	 */
	public void setRSSI(float RSSI) {
		this.rssi = RSSI;
	}

	/**
	 * Get the channel of this access point.
	 *
	 * @return the channel
	 */
	public int getChannel() {
		return channel;
	}

	/**
	 * Set the channel of the access point.
	 *
	 * @param channel
	 *            the channel
	 */
	public void setChannel(int channel) {
		this.channel = channel;
	}

	/**
	 * Get the security mode ({@link SecurityMode}).
	 *
	 * @return the security mode
	 */
	public SecurityMode getSecurityMode() {
		return securityMode;
	}

	/**
	 * Set the current security mode ({@link SecurityMode}).
	 *
	 * @param securityMode
	 *            the security mode
	 */
	public void setSecurityMode(SecurityMode securityMode) {
		this.securityMode = securityMode;
	}

	/**
	 * Get the Wi-Fi Protected Setup (WPS) modes supported by this access point.
	 *
	 * @return the supported WPS modes or an empty array if this access point doesn't support WPS
	 */
	public WPSMode[] getWPSCapabilities() {
		return wpsModes;
	}

	/**
	 * Set the supported WPS mode of the access point.
	 *
	 * @param modes
	 *            the WPS modes (must not be null)
	 */
	public void setWPSModes(WPSMode[] modes) {
		if (modes == null) {
			throw new IllegalArgumentException("WPS modes must not be null");
		}
		this.wpsModes = modes;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("AccessPointImpl [SSID=");
		builder.append(ssid);
		builder.append(", BSSID=");
		for (int i = 0; i < bssid.length - 1; i++) {
			builder.append(Integer.toHexString(bssid[i] & 0xFF));
			builder.append(':');
		}
		builder.append(Integer.toHexString(bssid[bssid.length - 1] & 0xFF));
		builder.append(", RSSI=");
		builder.append(rssi);
		builder.append(", channel=");
		builder.append(channel);
		builder.append(", securityMode=");
		builder.append(securityMode);
		builder.append(", wpsModes=");
		if (wpsModes.length == 0) {
			builder.append("none");
		} else {
			for (WPSMode mode : wpsModes) {
				builder.append(mode);
				builder.append(' ');
			}
		}
		builder.append("]");

		return builder.toString();
	}

}
