/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.vector.font;

import java.io.IOException;
import java.io.InputStream;

import ej.sni.SNI;
import ej.vector.font.map.PackedMap;

/**
 * A Vector font.
 */
public class VectorFont {

	private static final String FONT_ERROR = "Cannot load font file"; //$NON-NLS-1$

	private static final String NOT_IN_PATH = "Make sure the file is in application classpath and referenced in a '.resources.list' file."; //$NON-NLS-1$

	private static final String INVALID_FONT = "Make sure it is a valid TrueType font."; //$NON-NLS-1$

	private static final PackedMap<String, VectorFont> loadedFonts = new PackedMap<>();

	private final String path;

	private int faceReference;

	private static native int loadTrueTypeFont(byte[] fontPath);

	/**
	 * Gets an instance of {@link VectorFont} by its name.
	 *
	 * @param fontPath
	 *            the path of the font to get
	 * @return a singleton instance of the font
	 * @throws IOException
	 *             when the font cannot be loaded
	 */
	public static synchronized VectorFont getTrueTypeFont(String fontPath) throws IOException {
		if (loadedFonts.containsKey(fontPath)) {
			return loadedFonts.get(fontPath);
		}

		final VectorFont font = new VectorFont(fontPath);
		font.load();

		loadedFonts.put(fontPath, font);

		return font;
	}

	private VectorFont(String fontPath) {
		this.path = fontPath;
	}

	/**
	 * Gets the width of the given character with this font.<br>
	 * The width is the horizontal distance that would be occupied if the character was drawn using this font. It also
	 * includes the horizontal spaces between characters to separate them appropriately.
	 *
	 * @param size
	 *            the size of the font
	 * @param character
	 *            the character to measure
	 * @return the width taken by the given string.
	 */
	public int charWidth(float size, char character) {
		return charWidthNative(this.faceReference, size, character);
	}

	private native static int charWidthNative(int face, float size, char character);

	/**
	 * Gets the width of the given string with this font.<br>
	 * The width is the horizontal distance that would be occupied if the string was drawn using this font. It also
	 * includes the horizontal spaces between characters to separate them appropriately.
	 *
	 * @param size
	 *            the size of the font
	 * @param str
	 *            the string to measure
	 * @return the width taken by the given string.
	 */
	public int stringWidth(float size, String str) {
		return stringWidthNative(this.faceReference, size, str.toCharArray());
	}

	private native static int stringWidthNative(int face, float size, char[] str);

	/**
	 * Gets the reference of the loaded font face.
	 *
	 * @return an integer representing the face address in C native code.
	 */
	protected int getFaceReference() {
		return this.faceReference;
	}

	private void load() throws IOException {
		try (InputStream resource = VectorFont.class.getResourceAsStream(this.path)) {
			if (resource == null) {
				throw new IOException(getErrorMessage(NOT_IN_PATH));
			}
		}

		this.faceReference = loadTrueTypeFont(SNI.toCString(this.path));

		if (this.faceReference == 0) {
			throw new IOException(getErrorMessage(INVALID_FONT));
		}
	}

	private String getErrorMessage(String specificMessage) {
		final StringBuilder message = new StringBuilder(FONT_ERROR);
		message.append(" \"");
		message.append(this.path);
		message.append("\". ");
		message.append(specificMessage);
		return message.toString();
	}

}
