/*
 * Copyright (c) 1997, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package java.security;

import java.io.IOException;
import java.io.InputStream;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.util.Enumeration;

public class KeyStore {

	/*
	 * Constant to lookup in the Security properties file to determine the
	 * default keystore type. In the Security properties file, the default
	 * keystore type is given as: <pre> keystore.type=jks </pre>
	 */
	private static final String KEYSTORE_TYPE = "keystore.type";

	// The keystore type
	private final String type;

	// The provider implementation
	private final KeyStoreSpi keyStoreSpi;

	// Has this keystore been initialized (loaded)?
	private boolean initialized = false;

	private KeyStore(KeyStoreSpi keyStoreSpi, String type) {
		this.keyStoreSpi = keyStoreSpi;
		this.type = type;
	}

	public static KeyStore getInstance(String type) throws KeyStoreException {
		String spiImpl = System.getProperty("KeyStore." + type.toUpperCase());
		try {
			KeyStoreSpi keyStoreSpi = (KeyStoreSpi) Class.forName(spiImpl).newInstance();
			return new KeyStore(keyStoreSpi, type);
		} catch (Exception e) {
			throw new KeyStoreException(type + " not found", e);
		}
	}

	public final static String getDefaultType() {
		String kstype;
		kstype = System.getProperty(KEYSTORE_TYPE);
		if (kstype == null) {
			kstype = "jks";
		}
		return kstype;
	}

	public final String getType() {
		return this.type;
	}

	public final Key getKey(String alias, char[] password)
			throws KeyStoreException, NoSuchAlgorithmException, UnrecoverableKeyException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineGetKey(alias, password);
	}

	public final Certificate[] getCertificateChain(String alias) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineGetCertificateChain(alias);
	}

	public final Certificate getCertificate(String alias) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineGetCertificate(alias);
	}

	public final void setKeyEntry(String alias, byte[] key, Certificate[] chain) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		keyStoreSpi.engineSetKeyEntry(alias, key, chain);
	}

	public final void setCertificateEntry(String alias, Certificate cert) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		keyStoreSpi.engineSetCertificateEntry(alias, cert);
	}

	public final void deleteEntry(String alias) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		keyStoreSpi.engineDeleteEntry(alias);
	}

	public final Enumeration<String> aliases() throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineAliases();
	}

	public final boolean containsAlias(String alias) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineContainsAlias(alias);
	}

	public final int size() throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineSize();
	}

	public final boolean isKeyEntry(String alias) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineIsKeyEntry(alias);
	}

	public final boolean isCertificateEntry(String alias) throws KeyStoreException {
		if (!initialized) {
			throw new KeyStoreException("Uninitialized keystore");
		}
		return keyStoreSpi.engineIsCertificateEntry(alias);
	}

	public final void load(InputStream stream, char[] password)
			throws IOException, NoSuchAlgorithmException, CertificateException {
		keyStoreSpi.engineLoad(stream, password);
		initialized = true;
	}
}
