/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.microui;

import com.is2t.hil.HIL;

import ej.microui.display.LLUIDisplay;
import ej.microui.display.LLUIPainter.MicroUIGraphicsContext;
import ej.microui.display.LLUIPainter.MicroUIImage;
import ej.microui.display.LLUIPainter.MicroUIImageFormat;

/**
 * This class allows to draw an image with a rotation using an accelerated
 * algorithm.
 * <p>
 * Use same algorithm than BSP. Not specifically optimized for simulator.
 *
 * @see RotateCircular in MicroEJ application
 */
public class RotateCircular {

	private RotateCircular() {
		// static method only
	}

	public static int drawCircularImageWithArray(byte[] sd, byte[] img, int[] circularArea, int angle) {

		int ret = -1; // means not performed

		LLUIDisplay engine = LLUIDisplay.Instance;

		draw: {
			synchronized (engine) {

				MicroUIGraphicsContext gc = engine.mapMicroUIGraphicsContext(sd);
				MicroUIImage src = engine.mapMicroUIImage(img);

				if (src.getFormat() != MicroUIImageFormat.MICROUI_IMAGE_FORMAT_RGB565
						|| gc.getFormat() != MicroUIImageFormat.MICROUI_IMAGE_FORMAT_RGB565) {
					// not supported format
					break draw;
				}

				if (gc.getWidth() != src.getWidth() || gc.getHeight() != src.getHeight()) {
					// graphics contexts must have the same size
					break draw;
				}

				if (src.getWidth() != src.getWidth()) {
					// image must be a square
					break draw;
				}

				if (gc.requestDrawing()) {
					HIL.getInstance().refreshImmortalContent(circularArea);
					draw(gc, src, circularArea, angle);
					ret = 0; // drawing performed

					gc.setDrawingLimits(0, 0, gc.getWidth() - 1, gc.getHeight() - 1);
				}
			}
		}

		return ret;
	}

	private static void draw(MicroUIGraphicsContext gc, MicroUIImage src, int[] circularArea, int angle) {

		// retrieve src & dest characteristics
		int areaWidth = gc.getWidth();
		int areaHeight = gc.getHeight();

		// get angle in radian
		float angleRad = (angle * 3.1415927f) / 180.0f;
		float cosAngle = (float) Math.cos(angleRad);
		float sinAngle = (float) (-Math.sin(angleRad));

		// retrieve top-left coordinate on source (prevent to perform some multiplications in loops)
		int xCosMul = -areaWidth / 2;
		int yCosMul = -areaHeight / 2;
		float xSrcFromSrcOriginBase = xCosMul * cosAngle + yCosMul * sinAngle + areaWidth / 2;
		float ySrcFromSrcOriginBase = (-xCosMul) * sinAngle + yCosMul * cosAngle + areaHeight / 2;
		int circularAreaIndex = 0;

		for(int y = 0; y < areaHeight; y++)
		{
			float xSrcFromSrcOrigin = xSrcFromSrcOriginBase;
			float ySrcFromSrcOrigin = ySrcFromSrcOriginBase;

			// part 1: nothing to draw
			xSrcFromSrcOrigin += (cosAngle * circularArea[circularAreaIndex]);
			ySrcFromSrcOrigin -= (sinAngle * circularArea[circularAreaIndex]);
			int x = circularArea[circularAreaIndex];
			++circularAreaIndex;

			// part 2: have to draw
			for(; x < circularArea[circularAreaIndex]; x++)
			{
				if (((int)xSrcFromSrcOrigin >= 0) && ((int)ySrcFromSrcOrigin >= 0) && ((int)xSrcFromSrcOrigin < areaWidth) && ((int)ySrcFromSrcOrigin < areaHeight))
				{
					// retrieve source pixel color
					int pix = src.getImage().readPixel((int)xSrcFromSrcOrigin, (int)ySrcFromSrcOrigin);

					// draw into destination
					gc.getImage().drawPixel(x, y, pix);
				}
				// else: source pixel is out of bounds: nothing to draw on destination

				xSrcFromSrcOrigin += cosAngle;
				ySrcFromSrcOrigin -= sinAngle;
			}
			++circularAreaIndex;

			// part 3: nothing to draw
			xSrcFromSrcOrigin += (cosAngle * (circularArea[circularAreaIndex] - x + 1));
			ySrcFromSrcOrigin -= (sinAngle * (circularArea[circularAreaIndex] - x + 1));
			x = circularArea[circularAreaIndex];
			++circularAreaIndex;

			// part 4: have to draw
			for(; x < circularArea[circularAreaIndex]; x++)
			{
				if (((int)xSrcFromSrcOrigin >= 0) && ((int)ySrcFromSrcOrigin >= 0) && ((int)xSrcFromSrcOrigin < areaWidth) && ((int)ySrcFromSrcOrigin < areaHeight))
				{
					// retrieve source pixel color
					int pix = src.getImage().readPixel((int) xSrcFromSrcOrigin, (int) ySrcFromSrcOrigin);

					// draw into destination
					gc.getImage().drawPixel(x, y, pix);
				}
				// else: source pixel is out of bounds: nothing to draw on destination

				xSrcFromSrcOrigin += cosAngle;
				ySrcFromSrcOrigin -= sinAngle;
			}
			++circularAreaIndex;

			// go to next line
			xSrcFromSrcOriginBase += sinAngle;
			ySrcFromSrcOriginBase += cosAngle;
		}

	}
}
