/*
 * Java
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.verisilicon.vivante;

import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

/**
 *
 */
public class Matrix {
	private final List<MatrixOperation> ops;

	public Matrix() {
		ops = new ArrayList<>();
	}

	public int getHandle() {
		return this.hashCode();
	}

	private void add(MatrixOperation op) {
		ops.add(op);
	}

	private void clear() {
		ops.clear();
	}

	public void transform(Graphics2D g2d) {
		ListIterator<MatrixOperation> iter = ops.listIterator(ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			op.transform(g2d);
		}
	}

	public void transform(GeneralPath p) {
		ListIterator<MatrixOperation> iter = ops.listIterator(ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			op.transform(p);
		}
	}

	public void transform(Point2D source, Point2D destination) {
		ListIterator<MatrixOperation> iter = ops.listIterator(ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			op.transform(source, destination);
			source = destination;
		}
	}

	public Shape transform(Shape shape) {
		ListIterator<MatrixOperation> iter = ops.listIterator(ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			shape = op.transform(shape);
		}

		return shape;
	}

	/**
	 * See <code>vg_lite_identity</code> C function.
	 */
	public void identity() {
		// clear transformation
		clear();
		// apply a null translation in order to apply an identity transformation
		add(new OpTranslate(0, 0));
	}

	/**
	 * See <code>vg_lite_translate</code> C function.
	 */
	public void translate(float x, float y) {
		add(new OpTranslate(x, y));
	}

	/**
	 * See <code>vg_lite_scale</code> C function.
	 */
	public void scale(float scale_x, float scale_y) {
		add(new OpScale(scale_x, scale_y));
	}

	/**
	 * See <code>vg_lite_rotate</code> C function.
	 */
	public void rotate(float degrees) {
		add(new OpRotate(degrees));
	}

	/**
	 * See <code>vg_lite_perspective</code> C function.
	 */
	public static void perspective(float px, float py, Matrix matrix) {
		throw new UnsupportedOperationException();
	}

	private static interface MatrixOperation {
		public void transform(Graphics2D g2d);

		public void transform(GeneralPath p);

		public void transform(Point2D source, Point2D destination);

		public Shape transform(Shape shape);
	}

	private static class OpTranslate implements MatrixOperation {
		private final double _x;
		private final double _y;

		public OpTranslate(double x, double y) {
			_x = x;
			_y = y;
		}

		@Override
		public void transform(Graphics2D g2d) {
			g2d.translate(_x, _y);
		}

		@Override
		public void transform(GeneralPath p) {
			AffineTransform a = new AffineTransform();

			a.translate(_x, _y);
			p.transform(a);
		}

		@Override
		public void transform(Point2D source, Point2D destination) {
			AffineTransform a = new AffineTransform();

			a.translate(_x, _y);
			a.transform(source, destination);
		}

		@Override
		public Shape transform(Shape shape) {
			AffineTransform a = new AffineTransform();

			a.translate(_x, _y);
			return a.createTransformedShape(shape);
		}
	}

	private static class OpScale implements MatrixOperation {
		private final double _x;
		private final double _y;

		public OpScale(double x, double y) {
			_x = x;
			_y = y;
		}

		@Override
		public void transform(Graphics2D g2d) {
			g2d.scale(_x, _y);
		}

		@Override
		public void transform(GeneralPath p) {
			AffineTransform a = new AffineTransform();

			a.scale(_x, _y);
			p.transform(a);
		}

		@Override
		public void transform(Point2D source, Point2D destination) {
			AffineTransform a = new AffineTransform();

			a.scale(_x, _y);
			a.transform(source, destination);
		}

		@Override
		public Shape transform(Shape shape) {
			AffineTransform a = new AffineTransform();

			a.scale(_x, _y);
			return a.createTransformedShape(shape);
		}
	}

	private static class OpRotate implements MatrixOperation {
		private final double _t;

		public OpRotate(double t) {
			_t = t;
		}

		@Override
		public void transform(Graphics2D g2d) {
			g2d.rotate(_t);
		}

		@Override
		public void transform(GeneralPath p) {
			AffineTransform a = new AffineTransform();

			a.rotate(Math.toRadians(_t));
			p.transform(a);
		}

		@Override
		public void transform(Point2D source, Point2D destination) {
			AffineTransform a = new AffineTransform();

			a.rotate(Math.toRadians(_t));
			a.transform(source, destination);
		}

		@Override
		public Shape transform(Shape shape) {
			AffineTransform a = new AffineTransform();

			a.rotate(Math.toRadians(_t));
			return a.createTransformedShape(shape);
		}
	}
}
