/*
 * Java
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.verisilicon.vivante;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.LinearGradientPaint;
import java.awt.MultipleGradientPaint;
import java.awt.Paint;
import java.awt.RenderingHints;
import java.awt.geom.GeneralPath;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;

import com.is2t.hil.HIL;
import com.is2t.hil.NativeInterface;
import com.verisilicon.vivante.path.PathParserBase;

import ej.microui.display.LLUIDisplay;
import ej.microui.display.LLUIPainter.MicroUIGraphicsContext;

/**
 * This class provides a simulation of the Verisilicon VGLite APIs
 * on the front-panel
 */
public class Nvg_lite {

	public static final int VG_LITE_FILL_NON_ZERO = 0;
	public static final int VG_LITE_FILL_EVEN_ODD = 1;

	public static final int VG_LITE_BLEND_NONE = 0;
	public static final int VG_LITE_BLEND_SRC_OVER = 1;
	public static final int VG_LITE_BLEND_DST_OVER = 2;
	public static final int VG_LITE_BLEND_SRC_IN = 3;
	public static final int VG_LITE_BLEND_DST_IN = 4;
	public static final int VG_LITE_BLEND_SCREEN = 5;
	public static final int VG_LITE_BLEND_MULTIPLY = 6;
	public static final int VG_LITE_BLEND_ADDITIVE = 7;
	public static final int VG_LITE_BLEND_SUBTRACT = 8;

	private static final Point2D ORIGIN = new Point2D.Float(0f, 0f);
	private static final Point2D GRADIENT_END = new Point2D.Float(255f, 0f);

	public static int vg_lite_draw_path(byte[] sd, float pathLeft, float pathTop, float pathRight,
			float pathBottom,
			byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat, int fillRule,
			int matrixHandle,
			int blend, int color){
		try {
			return drawPath(sd, pathLeft, pathTop, pathRight, pathBottom, path, pathOffset, pathLength, pathQuality,
					pathFormat, fillRule, matrixHandle, blend, color);
		}catch (Throwable e) {
			// TODO manage error codes
			return -1;
		}
	}

	private static int drawPath(byte[] sd, float pathLeft, float pathTop, float pathRight, float pathBottom,
			byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat, int fillRule,
			int matrixHandle,
			int blend, int color){

		LLUIDisplay engine = LLUIDisplay.Instance;

		synchronized (engine) {

			MicroUIGraphicsContext gc = engine.mapMicroUIGraphicsContext(sd);
			if (gc.requestDrawing()) {

				// retrieve destination
				BufferedImage image = (BufferedImage) gc.getImage().getRAWImage();
				Graphics2D gc2d = image.createGraphics();

				// apply rendering hints
				RenderingHints rh = new RenderingHints(RenderingHints.KEY_ANTIALIASING,
						RenderingHints.VALUE_ANTIALIAS_ON);
				gc2d.setRenderingHints(rh);

				// apply color and composite
				setComposite(gc2d, blend);
				gc2d.setColor(new Color(getDisplayColor(color), true));

				// create VG path on user VG path
				refreshContent(path, pathOffset, pathLength);
				GeneralPath p = PathParserBase.get(path, pathOffset, pathLength, pathFormat);
				p.setWindingRule(getFillRule(fillRule));

				// apply VG path
				Matrix matrix = Nvg_lite_matrix.get(matrixHandle);
				matrix.transform(p);

				// Set the clip
				setClip(gc2d, gc, pathLeft, pathTop, pathRight, pathBottom, matrix);

				gc2d.fill(p);
			}
		}

		return 0;
	}

	private static int getFillRule(int fillRule) {
		switch (fillRule) {
		case VG_LITE_FILL_EVEN_ODD:
			return Path2D.WIND_EVEN_ODD;
		case VG_LITE_FILL_NON_ZERO:
		default:
			return Path2D.WIND_NON_ZERO;
		}
	}

	/**
	 * Clips the Graphics2D from the current MicroEJ GraphicsContext current clip
	 * and update drawing flush limits.
	 * <p>
	 * This code performs the same work than vg_lite library for embedded. It uses
	 * an orthogonal clip.
	 *
	 * @param gc2d
	 *            Graphics2D drawing object
	 * @param gc
	 *            MicroUI GraphicsContext where performing the drawing
	 * @param pathLeft
	 *            left bound of the path
	 * @param pathTop
	 *            top bound of the path
	 * @param pathRight
	 *            right bound of the path
	 * @param pathBottom
	 *            bottom bound of the path
	 * @param matrix
	 *            3x3 matrix that defines the transformation matrix of path
	 */
	private static void setClip(Graphics2D gc2d, MicroUIGraphicsContext gc, float pathLeft, float pathTop,
			float pathRight, float pathBottom, Matrix matrix) {

		Point2D min = new Point2D.Double();
		matrix.transform(new Point2D.Double(pathLeft, pathTop), min);
		Point2D max = (Point2D) min.clone();

		updateBounds(pathRight, pathTop, matrix, min, max);
		updateBounds(pathRight, pathBottom, matrix, min, max);
		updateBounds(pathLeft, pathBottom, matrix, min, max);
		updateBounds(pathRight, pathTop, matrix, min, max);

		int clipX1 = Math.max((int) min.getX(), gc.getClipX1());
		int clipY1 = Math.max((int) min.getY(), gc.getClipY1());
		int clipX2 = Math.min((int) max.getX(), gc.getClipX2());
		int clipY2 = Math.min((int) max.getY(), gc.getClipY2());

		gc2d.clipRect(clipX1, clipY1, clipX2 - clipX1 + 1, clipY2 - clipY1 + 1);

		// update rendering bounds
		gc.setDrawingLimits(clipX1, clipY1, clipX2, clipY2);
	}

	private static void updateBounds(float x, float y, Matrix matrix, Point2D min, Point2D max) {
		Point2D temp = new Point2D.Double();
		matrix.transform(new Point2D.Double(x, y), temp);

		double minX = min.getX();
		double minY = min.getY();
		double maxX = max.getX();
		double maxY = max.getY();

		if (temp.getX() < minX) {
			minX = temp.getX();
		}
		if (temp.getY() < minY) {
			minY = temp.getY();
		}
		if (temp.getX() > maxX) {
			maxX = temp.getX();
		}
		if (temp.getY() > maxY) {
			maxY = temp.getY();
		}

		min.setLocation(minX, minY);
		max.setLocation(maxX, maxY);
	}

	/**
	 * Refresh a rangee of an array; the array may be immortal or not.
	 *
	 * @param array
	 *            the array to refresh
	 * @param offset
	 *            the start index of the range.
	 * @param length
	 *            the length of the range.
	 * @return true if the array is immortal
	 */
	private static boolean refreshContent(byte[] array, int offset, int length) {
		NativeInterface hil = HIL.getInstance();
		boolean isImmortal = hil.isImmortal(array);
		if (isImmortal) {
			hil.refreshImmortalContent(array, offset, length);
		} else {
			hil.refreshContent(array, offset, length);
		}
		return isImmortal;
	}

	/**
	 * Refresh a rangee of an array; the array may be immortal or not.
	 *
	 * @param array
	 *            the array to refresh
	 * @param offset
	 *            the start index of the range.
	 * @param length
	 *            the length of the range.
	 * @return true if the array is immortal
	 */
	private static boolean refreshContent(int[] array, int offset, int length) {
		NativeInterface hil = HIL.getInstance();
		boolean isImmortal = hil.isImmortal(array);
		if (isImmortal) {
			hil.refreshImmortalContent(array, offset, length);
		} else {
			hil.refreshContent(array, offset, length);
		}
		return isImmortal;
	}

	/**
	 * Prints a warning message and sets the default graphics composite
	 * SRC_OVER.
	 *
	 * @param gc2d
	 *            the AWT graphics
	 * @param blendLabel
	 *            the name of the unsupported blend algorithm
	 */
	private static void setUnsupportedComposite(Graphics2D gc2d, String blendLabel) {
		System.err.println("Warning: unsupported blend mode: "
				+ blendLabel
				+ ", using BLEND_SRC_OVER instead");
		gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
	}

	/**
	 * Set graphics composite according user blending value.
	 *
	 * @param gc2d
	 *            the AWT graphics
	 * @param blend
	 *            the user blending value
	 * @throws IllegalArgumentException
	 *             if user blending value is not an expected value
	 */
	private static void setComposite(Graphics2D gc2d, int blend) {
		switch (blend) {
		case VG_LITE_BLEND_NONE:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC));
			break;
		case VG_LITE_BLEND_SRC_OVER:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
			break;
		case VG_LITE_BLEND_DST_OVER:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.DST_OVER));
			break;
		case VG_LITE_BLEND_SRC_IN:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_IN));
			break;
		case VG_LITE_BLEND_DST_IN:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.DST_IN));
			break;
		case VG_LITE_BLEND_SCREEN:
			setUnsupportedComposite(gc2d, "BLEND_SCREEN");
			break;
		case VG_LITE_BLEND_MULTIPLY:
			setUnsupportedComposite(gc2d, "BLEND_MULTIPLY");
			break;
		case VG_LITE_BLEND_ADDITIVE:
			setUnsupportedComposite(gc2d, "BLEND_ADDITIVE");
			break;
		default:
			throw new IllegalArgumentException("Unsupported blend mode: " + blend);
		}
	}

	public static int vg_lite_draw_gradient(byte[] gcArray, float pathLeft, float pathTop, float pathRight,
			float pathBottom, byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat,
			int fillRule, int matrixHandle, int gradientHandle, int blend) {
		try {
			return drawGradient(gcArray, pathLeft, pathTop, pathRight, pathBottom, path, pathOffset, pathLength,
					pathQuality,
					pathFormat, fillRule, matrixHandle, gradientHandle, blend);
		} catch (Throwable e) {
			// TODO manage error codes
			return -1;
		}
	}

	private static int drawGradient(byte[] gcArray, float pathLeft, float pathTop, float pathRight,
			float pathBottom, byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat,
			int fillRule, int matrixHandle, int gradientHandle, int blend) {

		LLUIDisplay engine = LLUIDisplay.Instance;
		synchronized (engine) {
			MicroUIGraphicsContext graphicContext = engine.mapMicroUIGraphicsContext(gcArray);
			if (graphicContext.requestDrawing()) {

				final Gradient gradient = Nvg_lite_gradient.get(gradientHandle);

				final Point2D gradientOrigin = new Point2D.Float();
				final Point2D gradientEnd = new Point2D.Float();

				final Matrix gradientMatrix = gradient.getMatrix();
				gradientMatrix.transform(ORIGIN, gradientOrigin);
				gradientMatrix.transform(GRADIENT_END, gradientEnd);

				final MultipleGradientPaint gradientPaint = new LinearGradientPaint(gradientOrigin, gradientEnd,
						gradient.getFractions(), gradient.getColors());

				Graphics2D g2d = (Graphics2D) ((BufferedImage) graphicContext.getImage().getRAWImage()).getGraphics();

				final Paint originalPaint = g2d.getPaint();
				g2d.setPaint(gradientPaint);

				// apply rendering hints
				RenderingHints rh = new RenderingHints(RenderingHints.KEY_ANTIALIASING,
						RenderingHints.VALUE_ANTIALIAS_ON);
				g2d.setRenderingHints(rh);
				setComposite(g2d, blend);

				// create VG path on user VG path
				refreshContent(path, pathOffset, pathLength);
				GeneralPath p = PathParserBase.get(path, pathOffset, pathLength, pathFormat);

				// apply VG path
				Matrix matrix = Nvg_lite_matrix.get(matrixHandle);
				matrix.transform(p);

				// Set the clip
				setClip(g2d, graphicContext, pathLeft, pathTop, pathRight, pathBottom, matrix);

				// paint
				g2d.fill(p);

				// reset paint
				g2d.setPaint(originalPaint);
			}
		}
		return 0;
	}

	public static int vg_lite_init_grad(int handle) {
		// Nothing to do on simulator
		return 0;
	}

	public static int vg_lite_set_grad(int handle, int count, int[] colors, int[] stops) {
		refreshContent(colors, 0, count);
		refreshContent(stops, 0, count);

		final Gradient gradient = Nvg_lite_gradient.get(handle);
		gradient.setFractions(stops);
		gradient.setColors(colors);
		return 0;
	}

	public static int vg_lite_update_grad(int handle) {
		// Nothing to do on simulator
		return 0;
	}

	public static int getDisplayColor(int color) {
		int alpha = (color & 0xFF000000);
		int rgb = (LLUIDisplay.Instance.convertARGBColorToColorToDraw(color) & 0x00FFFFFF);
		return (alpha | rgb);
	}
}
