/*
 * Java
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.vector.font;

import java.awt.Canvas;
import java.awt.Font;
import java.awt.FontFormatException;
import java.awt.FontMetrics;
import java.awt.Toolkit;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.is2t.hil.HIL;

/**
 * Native implementation of the equivalent class in Vector Font library.
 */
public class VectorFont {

	private static final int POINT_DPI = 72;

	/** Stores all the loaded fonts. A font is simply referenced by its index. */
	private static final List<Font> loadedFonts = new ArrayList<>();

	/**
	 * Loads a font given its path.
	 *
	 * This is the native implementation of a method from the FreeType library.
	 *
	 * @param fontPath
	 *            the path of the font in the application classpath in C string format
	 * @return a handle on the font
	 */
	public static int loadTrueTypeFont(byte[] fontPath) {
		HIL.getInstance().refreshImmortalContent(fontPath);
		try {
			// remove the trailing \0 (C string format) from path
			final String fontPathString = new String(fontPath, 0, fontPath.length - 1);
			final byte[] resourceData = HIL.getInstance().getResourceContent(fontPathString);
			final Font font = Font.createFont(Font.TRUETYPE_FONT, new ByteArrayInputStream(resourceData));
			loadedFonts.add(font);
		} catch (FontFormatException | IOException e) {
			// TODO handle errors
			e.printStackTrace();
		}

		// The handle will be the index in the table + 1
		// we avoid 0 because it is considered as NULL on board.
		return loadedFonts.size();
	}

	/**
	 * Gets a {@link Font} given its handle.
	 *
	 * The font must have been previously loaded.
	 *
	 * @param fontHandle
	 *            the font handle
	 * @param size
	 *            the font height
	 * @return the loaded {@link Font}
	 */
	public static final Font get(int fontHandle, float size) {
		try {
			// handle is index + 1
			Font font = loadedFonts.get(fontHandle - 1);

			if (font != null) {
				int pixelDpi = Toolkit.getDefaultToolkit().getScreenResolution();
				float pointSize = (float) POINT_DPI / pixelDpi;

				// get the font at the expected size
				font = font.deriveFont(size * pointSize);
			}


			return font;
		} catch (ArrayIndexOutOfBoundsException e) {
			// TODO handle errors
			throw e;
		}
	}

	public static int charWidthNative(int face, float size, char character) {
		final Font font = get(face, size);
		final FontMetrics metrics = new Canvas().getFontMetrics(font);

		return metrics.charWidth(character);
	}

	/**
	 * Gets the width of the given string with this font.<br>
	 * The width is the horizontal distance that would be occupied if the string was drawn using this font. It also
	 * includes the horizontal spaces between characters to separate them appropriately.
	 *
	 * @param gcArray
	 *            the graphic context
	 * @param face
	 *            the reference to the font
	 * @param size
	 *            the font size
	 * @param str
	 *            the string to measure as a char array
	 *
	 * @return the width taken by the given string.
	 */
	public static int stringWidthNative(int face, float size, char[] str) {
		NativeUtils.refreshContent(str, 0, str.length);

		final Font font = get(face, size);
		final FontMetrics metrics = getMetrics(font);

		int width = 0;

		for (char character : str) {
			width += metrics.charWidth(character);
		}

		// A workaround to avoid the last pixel column to be eaten by the string bounding box.
		return width + 1;
	}

	private static FontMetrics getMetrics(Font font) {
		return new Canvas().getFontMetrics(font);
	}
}
